"""Language registry with LanguageSpec definitions for all supported languages."""

import logging
import os
import threading
from dataclasses import dataclass
from typing import Optional


@dataclass
class LanguageSpec:
    """Specification for extracting symbols from a language's AST."""
    # tree-sitter language name (for tree-sitter-language-pack)
    ts_language: str

    # Node types that represent extractable symbols
    # Maps node_type -> symbol kind
    symbol_node_types: dict[str, str]

    # How to extract the symbol name from a node
    # Maps node_type -> child field name containing the name
    name_fields: dict[str, str]

    # How to extract parameters/signature beyond the name
    # Maps node_type -> child field name for parameters
    param_fields: dict[str, str]

    # Return type extraction (if language supports it)
    # Maps node_type -> child field name for return type
    return_type_fields: dict[str, str]

    # Docstring extraction strategy
    # "next_sibling_string" = Python (expression_statement after def)
    # "first_child_comment" = JS/TS (/** */ before function)
    # "preceding_comment" = Go/Rust/Java (// or /* */ before decl)
    docstring_strategy: str

    # Decorator/attribute node type (if any)
    decorator_node_type: Optional[str]

    # Node types that indicate nesting (methods inside classes)
    container_node_types: list[str]

    # Additional extraction: constants, type aliases
    constant_patterns: list[str]   # Node types for constants
    type_patterns: list[str]       # Node types for type definitions

    # If True, decorators are direct children of the declaration node (e.g. C#)
    # If False (default), decorators are preceding siblings (e.g. Python, Java)
    decorator_from_children: bool = False


# File extension to language mapping
LANGUAGE_EXTENSIONS = {
    ".py": "python",
    ".js": "javascript",
    ".jsx": "javascript",
    ".ts": "typescript",
    ".tsx": "tsx",
    ".go": "go",
    ".rs": "rust",
    ".java": "java",
    ".php": "php",
    ".dart": "dart",
    ".cs": "csharp",
    ".cshtml": "razor",
    ".razor": "razor",
    ".c": "c",
    ".h": "cpp",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".hpp": "cpp",
    ".hh": "cpp",
    ".hxx": "cpp",
    ".swift": "swift",
    ".ex": "elixir",
    ".exs": "elixir",
    ".rb": "ruby",
    ".rake": "ruby",
    ".pl": "perl",
    ".pm": "perl",
    ".t": "perl",
    ".gd": "gdscript",
    ".blade.php": "blade",
    ".al": "al",
    ".kt": "kotlin",
    ".kts": "kotlin",
    ".gleam": "gleam",
    ".sh": "bash",
    ".bash": "bash",
    ".nix": "nix",
    ".vue": "vue",
    ".ejs": "ejs",
    ".verse": "verse",
    ".lua": "lua",
    ".luau": "luau",
    ".erl": "erlang",
    ".hrl": "erlang",
    ".f90": "fortran",
    ".f95": "fortran",
    ".f03": "fortran",
    ".f08": "fortran",
    ".f": "fortran",
    ".for": "fortran",
    ".fpp": "fortran",
    ".sql": "sql",
    # Scala
    ".scala": "scala",
    ".sc": "scala",
    # Haskell
    ".hs": "haskell",
    ".lhs": "haskell",
    # Julia
    ".jl": "julia",
    # R
    ".r": "r",
    # CSS
    ".css": "css",
    # TOML
    ".toml": "toml",
    # Groovy
    ".groovy": "groovy",
    ".gradle": "groovy",
    # Objective-C
    ".m": "objc",
    ".mm": "objc",
    # Protobuf
    ".proto": "proto",
    # HCL / Terraform
    ".tf": "hcl",
    ".hcl": "hcl",
    ".tfvars": "hcl",
    # GraphQL
    ".graphql": "graphql",
    ".gql": "graphql",
    # Assembly (multi-dialect: WLA-DX, NASM, GAS, CA65, MASM, etc.)
    ".asm": "asm",
    ".s": "asm",
    ".S": "asm",
    ".inc": "asm",
    ".65816": "asm",
    ".z80": "asm",
    ".spc": "asm",
    ".6502": "asm",
    # AutoHotkey v2
    ".ahk": "autohotkey",
    ".ahk2": "autohotkey",
    # XML / XUL
    ".xml": "xml",
    ".xul": "xml",
    # YAML / Ansible (Ansible path heuristics handled in get_language_for_path)
    ".yaml": "yaml",
    ".yml": "yaml",
    # OpenAPI / Swagger (compound extensions; basenames handled in get_language_for_path)
    ".openapi.yaml": "openapi",
    ".openapi.yml": "openapi",
    ".openapi.json": "openapi",
    ".swagger.yaml": "openapi",
    ".swagger.yml": "openapi",
    ".swagger.json": "openapi",
}


# Python specification
PYTHON_SPEC = LanguageSpec(
    ts_language="python",
    symbol_node_types={
        "function_definition": "function",
        "class_definition": "class",
    },
    name_fields={
        "function_definition": "name",
        "class_definition": "name",
    },
    param_fields={
        "function_definition": "parameters",
    },
    return_type_fields={
        "function_definition": "return_type",
    },
    docstring_strategy="next_sibling_string",
    decorator_node_type="decorator",
    container_node_types=["class_definition"],
    constant_patterns=["assignment"],
    type_patterns=["type_alias_statement"],
)


# JavaScript specification
JAVASCRIPT_SPEC = LanguageSpec(
    ts_language="javascript",
    symbol_node_types={
        "function_declaration": "function",
        "class_declaration": "class",
        "method_definition": "method",
        "generator_function_declaration": "function",
    },
    name_fields={
        "function_declaration": "name",
        "class_declaration": "name",
        "method_definition": "name",
    },
    param_fields={
        "function_declaration": "parameters",
        "method_definition": "parameters",
        "arrow_function": "parameters",
    },
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=["class_declaration", "class"],
    constant_patterns=["lexical_declaration"],
    type_patterns=[],
)


# TSX specification (TypeScript with JSX — requires separate grammar)
TSX_SPEC = LanguageSpec(
    ts_language="tsx",
    symbol_node_types={
        "function_declaration": "function",
        "class_declaration": "class",
        "method_definition": "method",
        "interface_declaration": "type",
        "type_alias_declaration": "type",
        "enum_declaration": "type",
    },
    name_fields={
        "function_declaration": "name",
        "class_declaration": "name",
        "method_definition": "name",
        "interface_declaration": "name",
        "type_alias_declaration": "name",
        "enum_declaration": "name",
    },
    param_fields={
        "function_declaration": "parameters",
        "method_definition": "parameters",
        "arrow_function": "parameters",
    },
    return_type_fields={
        "function_declaration": "return_type",
        "method_definition": "return_type",
        "arrow_function": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="decorator",
    container_node_types=["class_declaration", "class"],
    constant_patterns=["lexical_declaration"],
    type_patterns=["interface_declaration", "type_alias_declaration", "enum_declaration"],
)


# TypeScript specification
TYPESCRIPT_SPEC = LanguageSpec(
    ts_language="typescript",
    symbol_node_types={
        "function_declaration": "function",
        "class_declaration": "class",
        "method_definition": "method",
        "interface_declaration": "type",
        "type_alias_declaration": "type",
        "enum_declaration": "type",
    },
    name_fields={
        "function_declaration": "name",
        "class_declaration": "name",
        "method_definition": "name",
        "interface_declaration": "name",
        "type_alias_declaration": "name",
        "enum_declaration": "name",
    },
    param_fields={
        "function_declaration": "parameters",
        "method_definition": "parameters",
        "arrow_function": "parameters",
    },
    return_type_fields={
        "function_declaration": "return_type",
        "method_definition": "return_type",
        "arrow_function": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="decorator",
    container_node_types=["class_declaration", "class"],
    constant_patterns=["lexical_declaration"],
    type_patterns=["interface_declaration", "type_alias_declaration", "enum_declaration"],
)


# Go specification
GO_SPEC = LanguageSpec(
    ts_language="go",
    symbol_node_types={
        "function_declaration": "function",
        "method_declaration": "method",
        "type_declaration": "type",
    },
    name_fields={
        "function_declaration": "name",
        "method_declaration": "name",
        "type_declaration": "name",
    },
    param_fields={
        "function_declaration": "parameters",
        "method_declaration": "parameters",
    },
    return_type_fields={
        "function_declaration": "result",
        "method_declaration": "result",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=["const_declaration"],
    type_patterns=["type_declaration"],
)


# Rust specification
RUST_SPEC = LanguageSpec(
    ts_language="rust",
    symbol_node_types={
        "function_item": "function",
        "struct_item": "type",
        "enum_item": "type",
        "trait_item": "type",
        "impl_item": "class",
        "type_item": "type",
    },
    name_fields={
        "function_item": "name",
        "struct_item": "name",
        "enum_item": "name",
        "trait_item": "name",
        "type_item": "name",
    },
    param_fields={
        "function_item": "parameters",
    },
    return_type_fields={
        "function_item": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="attribute_item",
    container_node_types=["impl_item", "trait_item"],
    constant_patterns=["const_item", "static_item"],
    type_patterns=["struct_item", "enum_item", "trait_item", "type_item"],
)


# Java specification
JAVA_SPEC = LanguageSpec(
    ts_language="java",
    symbol_node_types={
        "method_declaration": "method",
        "constructor_declaration": "method",
        "class_declaration": "class",
        "interface_declaration": "type",
        "enum_declaration": "type",
    },
    name_fields={
        "method_declaration": "name",
        "constructor_declaration": "name",
        "class_declaration": "name",
        "interface_declaration": "name",
        "enum_declaration": "name",
    },
    param_fields={
        "method_declaration": "parameters",
        "constructor_declaration": "parameters",
    },
    return_type_fields={
        "method_declaration": "type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="marker_annotation",
    container_node_types=["class_declaration", "interface_declaration", "enum_declaration"],
    constant_patterns=["field_declaration"],
    type_patterns=["interface_declaration", "enum_declaration"],
)


# PHP specification
PHP_SPEC = LanguageSpec(
    ts_language="php",
    symbol_node_types={
        "function_definition": "function",
        "class_declaration": "class",
        "method_declaration": "method",
        "interface_declaration": "type",
        "trait_declaration": "type",
        "enum_declaration": "type",
    },
    name_fields={
        "function_definition": "name",
        "class_declaration": "name",
        "method_declaration": "name",
        "interface_declaration": "name",
        "trait_declaration": "name",
        "enum_declaration": "name",
    },
    param_fields={
        "function_definition": "parameters",
        "method_declaration": "parameters",
    },
    return_type_fields={
        "function_definition": "return_type",
        "method_declaration": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="attribute",  # PHP 8 #[Attribute] syntax
    container_node_types=["class_declaration", "trait_declaration", "interface_declaration"],
    constant_patterns=["const_declaration"],
    type_patterns=["interface_declaration", "trait_declaration", "enum_declaration"],
)


# Dart specification
DART_SPEC = LanguageSpec(
    ts_language="dart",
    symbol_node_types={
        "function_signature": "function",
        "class_definition": "class",
        "mixin_declaration": "class",
        "enum_declaration": "type",
        "extension_declaration": "class",
        "method_signature": "method",
        "type_alias": "type",
    },
    name_fields={
        "function_signature": "name",
        "class_definition": "name",
        "enum_declaration": "name",
        "extension_declaration": "name",
        # mixin_declaration, method_signature, type_alias: special-cased in extractor
    },
    param_fields={
        "function_signature": "parameters",
    },
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type="annotation",
    container_node_types=["class_definition", "mixin_declaration", "extension_declaration"],
    constant_patterns=[],
    type_patterns=["type_alias", "enum_declaration"],
)


# C# specification
CSHARP_SPEC = LanguageSpec(
    ts_language="csharp",
    symbol_node_types={
        "class_declaration": "class",
        "record_declaration": "class",
        "interface_declaration": "type",
        "enum_declaration": "type",
        "struct_declaration": "type",
        "delegate_declaration": "type",
        "method_declaration": "method",
        "constructor_declaration": "method",
        "property_declaration": "constant",
        "field_declaration": "constant",
        "event_field_declaration": "constant",
        "event_declaration": "constant",
        "destructor_declaration": "method",
    },
    name_fields={
        "class_declaration": "name",
        "record_declaration": "name",
        "interface_declaration": "name",
        "enum_declaration": "name",
        "struct_declaration": "name",
        "delegate_declaration": "name",
        "method_declaration": "name",
        "constructor_declaration": "name",
        "property_declaration": "name",
        "event_declaration": "name",
        "destructor_declaration": "name",
    },
    param_fields={
        "method_declaration": "parameters",
        "constructor_declaration": "parameters",
        "delegate_declaration": "parameters",
    },
    return_type_fields={
        "method_declaration": "returns",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="attribute_list",
    decorator_from_children=True,
    container_node_types=["class_declaration", "struct_declaration", "record_declaration", "interface_declaration"],
    constant_patterns=[],
    type_patterns=["interface_declaration", "enum_declaration", "struct_declaration", "delegate_declaration", "record_declaration"],
)


# Razor / ASP.NET views + Blazor components specification
# NOTE: .cshtml (MVC views) and .razor (Blazor components) are both mixed-language
# documents containing Razor directives, HTML markup, optional <script>/<style>
# blocks, and embedded C# in @functions/@code regions.  Extraction is handled by
# _parse_razor_symbols() in extractor.py, which delegates subregions to the
# existing C#/JS parsers and emits lightweight symbols for HTML ids, external
# scripts, style blocks, and (for Blazor) @page routes and @inject directives.
RAZOR_SPEC = LanguageSpec(
    ts_language="html",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# C specification
C_SPEC = LanguageSpec(
    ts_language="c",
    symbol_node_types={
        "function_definition": "function",
        "struct_specifier": "type",
        "enum_specifier": "type",
        "union_specifier": "type",
        "type_definition": "type",
    },
    name_fields={
        "function_definition": "declarator",
        "struct_specifier": "name",
        "enum_specifier": "name",
        "union_specifier": "name",
        "type_definition": "declarator",
    },
    param_fields={
        "function_definition": "declarator",
    },
    return_type_fields={
        "function_definition": "type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=["preproc_def"],
    type_patterns=["type_definition", "enum_specifier", "struct_specifier", "union_specifier"],
)


# Swift specification
# Note: tree-sitter-swift uses class_declaration for class/struct/enum/extension;
# the declaration_kind child field ("class"/"struct"/"enum"/"extension") disambiguates
# at the source level but all map to "class" here for uniform treatment.
# Attributes (@discardableResult etc.) live inside a modifiers child node rather
# than as preceding siblings, so decorator extraction is not supported in this spec.
SWIFT_SPEC = LanguageSpec(
    ts_language="swift",
    symbol_node_types={
        "function_declaration": "function",
        "class_declaration": "class",    # covers class, struct, enum, extension
        "protocol_declaration": "type",
        "typealias_declaration": "type",
        "init_declaration": "method",
        "deinit_declaration": "method",
        "property_declaration": "constant",
    },
    name_fields={
        "function_declaration": "name",  # simple_identifier child
        "class_declaration": "name",     # type_identifier child
        "protocol_declaration": "name",  # type_identifier child
        "typealias_declaration": "name", # user_type child
        "init_declaration": "name",      # "init" keyword token
        "deinit_declaration": "name",    # "deinit" keyword token
        "property_declaration": "name",  # pattern child
    },
    param_fields={},  # Swift params are unnamed children; signature captured via source range
    return_type_fields={},  # return type shares field "name" with function identifier
    docstring_strategy="preceding_comment",  # /// and /* */ doc comments
    decorator_node_type=None,
    container_node_types=["class_declaration", "protocol_declaration"],
    constant_patterns=[],  # property_declaration handled via symbol_node_types
    type_patterns=["protocol_declaration", "typealias_declaration"],
)


# C++ specification
CPP_SPEC = LanguageSpec(
    ts_language="cpp",
    symbol_node_types={
        "class_specifier": "class",
        "struct_specifier": "type",
        "union_specifier": "type",
        "enum_specifier": "type",
        "type_definition": "type",
        "alias_declaration": "type",
        "function_definition": "function",
        "declaration": "function",
        "field_declaration": "function",
    },
    name_fields={
        "class_specifier": "name",
        "struct_specifier": "name",
        "union_specifier": "name",
        "enum_specifier": "name",
        "type_definition": "declarator",
        "alias_declaration": "name",
        "function_definition": "declarator",
        "declaration": "declarator",
        "field_declaration": "declarator",
    },
    param_fields={
        "function_definition": "declarator",
        "declaration": "declarator",
        "field_declaration": "declarator",
    },
    return_type_fields={
        "function_definition": "type",
        "declaration": "type",
        "field_declaration": "type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=["class_specifier", "struct_specifier", "union_specifier"],
    constant_patterns=["preproc_def"],
    type_patterns=["class_specifier", "struct_specifier", "union_specifier", "enum_specifier", "type_definition", "alias_declaration"],
)


# Elixir specification
# NOTE: Elixir's tree-sitter grammar is homoiconic — all constructs (defmodule,
# def, defp, defmacro, @doc, @type, etc.) are represented as generic `call` or
# `unary_operator` nodes. Custom extraction is performed in extractor.py via
# _parse_elixir_symbols(); the fields below are intentionally empty.
ELIXIR_SPEC = LanguageSpec(
    ts_language="elixir",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="elixir",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Perl specification
PERL_SPEC = LanguageSpec(
    ts_language="perl",
    symbol_node_types={
        "subroutine_declaration_statement": "function",
        "package_statement": "class",
    },
    name_fields={
        "subroutine_declaration_statement": "name",
        "package_statement": "name",
    },
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=["use_statement"],
    type_patterns=[],
)


# Ruby specification
RUBY_SPEC = LanguageSpec(
    ts_language="ruby",
    symbol_node_types={
        "method": "function",           # top-level → function; inside class/module → method
        "singleton_method": "function", # def self.foo → always has class parent → method
        "class": "class",
        "module": "type",
    },
    name_fields={
        "method": "name",
        "singleton_method": "name",
        "class": "name",
        "module": "name",
    },
    param_fields={
        "method": "parameters",
        "singleton_method": "parameters",
    },
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=["class", "module"],
    constant_patterns=[],
    type_patterns=["module"],
)


# GDScript specification (Godot 4)
# GDScript is Python-like; tree-sitter-gdscript exposes named fields
# for name, parameters, and return_type on function_definition nodes.
# Annotations (@export, @onready, @tool) appear as preceding `annotation`
# siblings before the declaration they decorate.
GDSCRIPT_SPEC = LanguageSpec(
    ts_language="gdscript",
    symbol_node_types={
        "function_definition": "function",
        "class_definition": "class",
        "signal_statement": "function",
        "enum_definition": "type",
    },
    name_fields={
        "function_definition": "name",
        "class_definition": "name",
        "signal_statement": "name",
        "enum_definition": "name",
    },
    param_fields={
        "function_definition": "parameters",
        "signal_statement": "parameters",
    },
    return_type_fields={
        "function_definition": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="annotation",
    container_node_types=["class_definition"],
    constant_patterns=["const_statement"],
    type_patterns=["enum_definition"],
)


# Blade (Laravel) specification
# NOTE: No tree-sitter grammar is available for Blade templates.
# Symbol extraction is performed by _parse_blade_symbols() in extractor.py
# via regex scanning for Blade directives (@section, @component, @extends, etc.).
# The fields below are intentionally empty.
BLADE_SPEC = LanguageSpec(
    ts_language="blade",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# AL (Business Central) specification
# NOTE: No tree-sitter grammar is available for AL.
# Symbol extraction is performed by _parse_al_symbols() in extractor.py
# via regex scanning for object declarations, procedures, triggers, and fields.
# The fields below are intentionally empty.
AL_SPEC = LanguageSpec(
    ts_language="al",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Kotlin specification
# NOTE: Kotlin's tree-sitter grammar exposes no named field accessors for names,
# parameters, or bodies. All extraction is handled via special-cases in extractor.py
# that walk children by node type (simple_identifier / type_identifier / function_body).
KOTLIN_SPEC = LanguageSpec(
    ts_language="kotlin",
    symbol_node_types={
        "class_declaration": "class",     # class, interface, enum class, data class
        "object_declaration": "class",    # object declarations (singletons)
        "function_declaration": "function",
        "type_alias": "type",
    },
    name_fields={},     # Names extracted via special-case in extractor.py
    param_fields={},    # Parameters captured via source range in _build_signature
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,  # Annotations live inside modifiers node; captured in signature
    container_node_types=["class_declaration", "object_declaration"],
    constant_patterns=["property_declaration"],
    type_patterns=["type_alias", "class_declaration"],
)


# Gleam specification
GLEAM_SPEC = LanguageSpec(
    ts_language="gleam",
    symbol_node_types={
        "function": "function",
        "type_definition": "type",
        "type_alias": "type",
        "constant": "constant",
    },
    name_fields={
        "function": "name",    # identifier field
        "constant": "name",    # identifier field
        # type_definition and type_alias: name via type_name child, special-cased in extractor.py
    },
    param_fields={
        "function": "parameters",
    },
    return_type_fields={
        "function": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=["constant"],
    type_patterns=["type_definition", "type_alias"],
)


# Bash specification
BASH_SPEC = LanguageSpec(
    ts_language="bash",
    symbol_node_types={
        "function_definition": "function",
    },
    name_fields={
        "function_definition": "name",
    },
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=["declaration_command"],  # readonly / declare -r
    type_patterns=[],
)


# Nix specification
# NOTE: Nix is an expression language; all constructs are `binding` nodes inside
# binding_set children of let_expression or attrset_expression. Custom extraction
# is performed in extractor.py via _parse_nix_symbols(). Fields below are empty.
NIX_SPEC = LanguageSpec(
    ts_language="nix",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# EJS (Embedded JavaScript Templates) specification
# NOTE: No tree-sitter grammar exists for EJS. Extraction is handled by
# _parse_ejs_symbols() in extractor.py via regex, which pulls JS function
# definitions from <% %> scriptlet blocks and emits a synthetic "template"
# symbol per file to ensure the file is always stored for text search.
EJS_SPEC = LanguageSpec(
    ts_language="ejs",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Vue SFC specification
# NOTE: Vue Single-File Components are parsed by _parse_vue_symbols() in
# extractor.py, which extracts the <script>/<script setup> block and re-parses
# it as JavaScript or TypeScript (detected from the lang="ts" attribute).
VUE_SPEC = LanguageSpec(
    ts_language="vue",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Verse (UEFN) specification
# NOTE: No tree-sitter grammar exists for Epic's Verse language.
# Symbol extraction is performed by _parse_verse_symbols() in extractor.py
# using regex-based parsing (same approach as Blade).
#
# Primary use case: token-efficient lookup of UEFN API digest files.
# The three standard Verse digest files total ~800KB / ~200k tokens:
#   Fortnite.digest.verse    587KB  3,608 symbols  ~147k tokens
#   Verse.digest.verse       125KB    622 symbols   ~31k tokens
#   UnrealEngine.digest.verse 91KB    326 symbols   ~23k tokens
#
# With jcodemunch indexing, a single symbol lookup costs ~94 tokens
# instead of loading the full file (~147k tokens) — a 99.9% reduction.
# A search returning 10 signature matches costs ~130 tokens.
#
# The LanguageSpec fields below are intentionally empty — all extraction
# logic lives in _parse_verse_symbols() which handles containers, methods,
# extension methods, variables, constants, docstrings, and decorators.
VERSE_SPEC = LanguageSpec(
    ts_language="verse",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Fortran specification
# NOTE: Fortran's tree-sitter grammar uses a translation_unit root with
# function/subroutine/module/program top-level nodes.  function_statement
# and subroutine_statement expose named 'name' and 'parameters' fields.
# Module-contained procedures live inside internal_procedures nodes.
# All extraction logic is in _parse_fortran_symbols() in extractor.py.
FORTRAN_SPEC = LanguageSpec(
    ts_language="fortran",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Erlang specification
# NOTE: Erlang's tree-sitter grammar represents all top-level constructs as
# distinct node types (fun_decl, type_alias, opaque, record_decl, pp_define).
# Multi-clause functions produce one fun_decl per clause; deduplication by
# (name, arity) is handled in _parse_erlang_symbols() in extractor.py.
# Named fields are not used (the grammar doesn't expose them uniformly), so
# all fields below are intentionally empty — extraction logic lives in
# _parse_erlang_symbols().
ERLANG_SPEC = LanguageSpec(
    ts_language="erlang",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Lua specification
# NOTE: Lua uses a single `function_declaration` node type for all named
# functions — local, module.method (dot_index_expression), and OOP methods
# (method_index_expression). Custom extraction is performed by
# _parse_lua_symbols() in extractor.py, which handles name resolution and
# method vs. function classification. Fields below are intentionally empty.
LUA_SPEC = LanguageSpec(
    ts_language="lua",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Luau specification (Roblox)
# NOTE: Luau is Roblox's typed superset of Lua. The tree-sitter-luau grammar
# uses the same ``function_declaration`` node type as Lua, with the same
# ``name``, ``parameters``, and ``body`` fields, plus Luau-specific constructs:
# - ``type_definition`` for ``type Foo = ...`` and ``export type Foo = ...``
# - Typed parameters (``param: Type``) inside ``parameter`` children
# - Return type annotations after the closing ``)``
# Custom extraction is performed by _parse_luau_symbols() in extractor.py.
LUAU_SPEC = LanguageSpec(
    ts_language="luau",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# SQL specification
# NOTE: The derekstride/tree-sitter-sql grammar has no named field accessors.
# Names live in positional children (object_reference → identifier, or direct
# identifier). CREATE PROCEDURE and CREATE TRIGGER produce ERROR nodes and are
# not supported by this grammar. All extraction logic is in _parse_sql_symbols()
# in extractor.py.
SQL_SPEC = LanguageSpec(
    ts_language="sql",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Scala specification
SCALA_SPEC = LanguageSpec(
    ts_language="scala",
    symbol_node_types={
        "class_definition": "class",
        "object_definition": "class",
        "trait_definition": "type",
        "enum_definition": "type",
        "function_definition": "function",
    },
    name_fields={
        "class_definition": "name",
        "object_definition": "name",
        "trait_definition": "name",
        "enum_definition": "name",
        "function_definition": "name",
    },
    param_fields={
        "function_definition": "parameters",
    },
    return_type_fields={
        "function_definition": "return_type",
    },
    docstring_strategy="preceding_comment",
    decorator_node_type="annotation",
    container_node_types=["class_definition", "object_definition", "trait_definition"],
    constant_patterns=["val_definition", "var_definition"],
    type_patterns=["trait_definition", "enum_definition"],
)


# Haskell specification
# NOTE: Haskell's tree-sitter grammar represents declarations as complex nested
# nodes without standard named fields. Full extraction is deferred to a future
# custom parser. Files are indexed for text search; symbol extraction is minimal.
HASKELL_SPEC = LanguageSpec(
    ts_language="haskell",
    symbol_node_types={
        "function": "function",
        "data_type": "type",
        "type_synon": "type",
        "newtype": "type",
        "class": "type",
    },
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=["data_type", "type_synon", "newtype"],
)


# Julia specification
# NOTE: Julia's tree-sitter grammar nests function names inside signature nodes
# rather than exposing them as direct named fields. Custom extraction is handled
# by _parse_julia_symbols() in extractor.py. Fields below are intentionally empty.
JULIA_SPEC = LanguageSpec(
    ts_language="julia",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# R specification
# NOTE: R functions are values assigned to names (e.g. foo <- function(x) {...}).
# The generic extractor cannot handle this pattern. Files are indexed for text
# search; a custom extractor will be added in a future pass.
R_SPEC = LanguageSpec(
    ts_language="r",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# CSS specification
# NOTE: CSS rule sets use selectors as names. Symbol extraction is deferred to
# a future custom parser. Files are indexed for text search.
CSS_SPEC = LanguageSpec(
    ts_language="css",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# TOML specification
# NOTE: TOML tables are the closest analogue to symbols. Custom extraction
# deferred. Files are indexed for text search.
TOML_SPEC = LanguageSpec(
    ts_language="toml",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Groovy specification
# NOTE: tree-sitter-groovy uses a low-level grammar (command/unit/block/func nodes)
# rather than Java-style named declarations. Custom extraction is in extractor.py.
GROOVY_SPEC = LanguageSpec(
    ts_language="groovy",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Objective-C specification
# NOTE: ObjC class interface/implementation and method declarations use
# non-standard selector-based naming. Custom extraction is in extractor.py.
OBJC_SPEC = LanguageSpec(
    ts_language="objc",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Protocol Buffers specification
# NOTE: Custom extraction in extractor.py handles message/service/rpc/enum
# name resolution from child nodes.
PROTO_SPEC = LanguageSpec(
    ts_language="proto",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# HCL / Terraform specification
# NOTE: HCL blocks (resource, data, module, variable, output, locals) are
# extracted as symbols by _parse_hcl_symbols() in extractor.py.
HCL_SPEC = LanguageSpec(
    ts_language="hcl",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# GraphQL specification
# NOTE: GraphQL type/query/mutation/fragment definitions are extracted by
# _parse_graphql_symbols() in extractor.py.
GRAPHQL_SPEC = LanguageSpec(
    ts_language="graphql",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Assembly language specification
# NOTE: No tree-sitter grammar covers the breadth of assembler dialects used
# in retro and embedded development (WLA-DX, NASM, GAS, CA65, etc.).  Symbol
# extraction is handled entirely by _parse_asm_symbols() in extractor.py using
# regex line-scanning.  Supports labels, sections, macros, constants (.define,
# .def, .set, .equ, %define, equ), structs, enums, and procedures across
# multiple assembler syntaxes.  Fields below are intentionally empty;
# ts_language is a placeholder never used.
ASM_SPEC = LanguageSpec(
    ts_language="asm",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# AutoHotkey v2 specification
# NOTE: AutoHotkey is not available in tree-sitter-language-pack and has no
# published standalone tree-sitter Python binding.  Symbol extraction is
# handled entirely by _parse_autohotkey_symbols() in extractor.py using
# regex line-scanning with brace-depth tracking.  Fields below are
# intentionally empty; ts_language is a placeholder never used.
AHK_SPEC = LanguageSpec(
    ts_language="autohotkey",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# XML / XUL specification
# NOTE: XML and XUL (Mozilla's XML User Interface Language) share the same
# tree-sitter-xml grammar. XUL is a strict XML superset — same node types,
# same attributes. Custom extraction is performed by _parse_xml_symbols()
# in extractor.py, which extracts:
#   - Document root element (e.g. <window>, <page>) → type symbol
#   - Elements with id attributes (e.g. <textbox id="search">) → constant symbols
#   - <script src="..."> references → function symbols
# Fields below are intentionally empty — all extraction logic lives in
# _parse_xml_symbols().
XML_SPEC = LanguageSpec(
    ts_language="xml",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


YAML_SPEC = LanguageSpec(
    ts_language="yaml",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


ANSIBLE_SPEC = LanguageSpec(
    ts_language="yaml",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# OpenAPI / Swagger specification
# NOTE: Parsed by _parse_openapi_symbols() in extractor.py using yaml/json.
# File detection uses compound extensions (.openapi.yaml, .swagger.json, …)
# plus well-known basenames (openapi.yaml, swagger.json) handled in
# get_language_for_path(). ts_language is unused — extraction is dict-based.
OPENAPI_SPEC = LanguageSpec(
    ts_language="yaml",
    symbol_node_types={},
    name_fields={},
    param_fields={},
    return_type_fields={},
    docstring_strategy="preceding_comment",
    decorator_node_type=None,
    container_node_types=[],
    constant_patterns=[],
    type_patterns=[],
)


# Language registry
LANGUAGE_REGISTRY = {
    "python": PYTHON_SPEC,
    "javascript": JAVASCRIPT_SPEC,
    "typescript": TYPESCRIPT_SPEC,
    "tsx": TSX_SPEC,
    "go": GO_SPEC,
    "rust": RUST_SPEC,
    "java": JAVA_SPEC,
    "php": PHP_SPEC,
    "dart": DART_SPEC,
    "csharp": CSHARP_SPEC,
    "razor": RAZOR_SPEC,
    "c": C_SPEC,
    "swift": SWIFT_SPEC,
    "cpp": CPP_SPEC,
    "elixir": ELIXIR_SPEC,
    "ruby": RUBY_SPEC,
    "perl": PERL_SPEC,
    "gdscript": GDSCRIPT_SPEC,
    "blade": BLADE_SPEC,
    "al": AL_SPEC,
    "kotlin": KOTLIN_SPEC,
    "gleam": GLEAM_SPEC,
    "bash": BASH_SPEC,
    "nix": NIX_SPEC,
    "vue": VUE_SPEC,
    "ejs": EJS_SPEC,
    "verse": VERSE_SPEC,
    "lua": LUA_SPEC,
    "luau": LUAU_SPEC,
    "erlang": ERLANG_SPEC,
    "fortran": FORTRAN_SPEC,
    "sql": SQL_SPEC,
    "scala": SCALA_SPEC,
    "haskell": HASKELL_SPEC,
    "julia": JULIA_SPEC,
    "r": R_SPEC,
    "css": CSS_SPEC,
    "toml": TOML_SPEC,
    "groovy": GROOVY_SPEC,
    "objc": OBJC_SPEC,
    "proto": PROTO_SPEC,
    "hcl": HCL_SPEC,
    "graphql": GRAPHQL_SPEC,
    "autohotkey": AHK_SPEC,
    "asm": ASM_SPEC,
    "xml": XML_SPEC,
    "yaml": YAML_SPEC,
    "ansible": ANSIBLE_SPEC,
    "openapi": OPENAPI_SPEC,
}

logger = logging.getLogger(__name__)

_APPLIED_EXTENSIONS = False
_EXTENSIONS_LOCK = threading.Lock()


_OPENAPI_BASENAMES = frozenset({
    "openapi.yaml", "openapi.yml", "openapi.json",
    "swagger.yaml", "swagger.yml", "swagger.json",
})

_ANSIBLE_PLAYBOOK_DIRS = frozenset({"playbooks", "playbook"})
_ANSIBLE_ROLE_DIR_SEGMENTS = frozenset({"tasks", "handlers", "vars", "defaults", "meta"})
_ANSIBLE_SPECIAL_DIRS = frozenset({"group_vars", "host_vars"})
_ANSIBLE_BASENAMES = frozenset({
    "site.yml", "site.yaml",
    "requirements.yml", "requirements.yaml",
    "galaxy.yml", "galaxy.yaml",
})


def _looks_like_ansible_path(path: str) -> bool:
    """Best-effort path heuristics for Ansible YAML files."""
    lower = path.lower().replace("\\", "/")
    base = os.path.basename(lower)
    parts = [part for part in lower.split("/") if part]
    if not base.endswith((".yml", ".yaml")):
        return False

    if base in _ANSIBLE_BASENAMES:
        return True
    if any(marker in parts for marker in _ANSIBLE_PLAYBOOK_DIRS):
        return True
    if any(marker in parts for marker in _ANSIBLE_SPECIAL_DIRS):
        return True

    for idx, part in enumerate(parts):
        if part != "roles" or idx + 2 >= len(parts):
            continue
        if parts[idx + 2] in _ANSIBLE_ROLE_DIR_SEGMENTS:
            return True
    return False


def _apply_extra_extensions() -> None:
    """Apply extra_extensions from config to LANGUAGE_EXTENSIONS.

    Respects config.get("extra_extensions") which includes env var fallback
    (both JSON and legacy comma-separated formats) applied at startup.

    Lazy — runs once on first access (cached via _APPLIED_EXTENSIONS flag).
    Thread-safe via _EXTENSIONS_LOCK.
    """
    global _APPLIED_EXTENSIONS
    with _EXTENSIONS_LOCK:
        if _APPLIED_EXTENSIONS:
            return

        from .. import config as _cfg

        extra_extensions = _cfg.get("extra_extensions", {})
        for ext, lang in extra_extensions.items():
            if not ext or not lang:
                logger.warning("extra_extensions: empty extension or language %r:%r — skipped", ext, lang)
                continue
            if lang not in LANGUAGE_REGISTRY:
                logger.warning("extra_extensions: unknown language %r for extension %r — skipped", lang, ext)
                continue
            LANGUAGE_EXTENSIONS[ext] = lang

        _APPLIED_EXTENSIONS = True


def get_language_extensions() -> dict[str, str]:
    """Return LANGUAGE_EXTENSIONS with extra_extensions applied (lazy, cached)."""
    _apply_extra_extensions()
    return LANGUAGE_EXTENSIONS


def get_language_for_path(path: str) -> "Optional[str]":
    """Return the language name for a file path, handling compound extensions.

    Check order:
    1. Well-known OpenAPI/Swagger basenames (openapi.yaml, swagger.json, …).
    2. Ansible path heuristics for YAML inventory/playbook files.
    3. Compound suffixes (e.g. ``.blade.php``, ``.openapi.yaml``).
    4. Last extension (e.g. ``.php``).
    """
    _apply_extra_extensions()
    import os as _os
    lower = path.lower()
    base = _os.path.basename(lower)
    # 1. Basename match for OpenAPI sentinel files
    if base in _OPENAPI_BASENAMES:
        return "openapi"
    # 2. Path heuristics for Ansible YAML files
    if _looks_like_ansible_path(lower):
        return "ansible"
    # 3. Compound extension (e.g. ".blade.php", ".openapi.yaml")
    first_dot = base.find(".")
    if first_dot != -1:
        compound = base[first_dot:]
        if compound in LANGUAGE_EXTENSIONS:
            return LANGUAGE_EXTENSIONS[compound]
    # 4. Simple extension
    _, ext = _os.path.splitext(lower)
    return LANGUAGE_EXTENSIONS.get(ext)



"""Tests for YAML and Ansible parsing support."""

from jcodemunch_mcp.parser import parse_file
from jcodemunch_mcp.parser.languages import get_language_for_path


GENERIC_YAML_SOURCE = """
services:
  api:
    image: example/api:latest
    ports:
      - "8080:8080"
feature_flags:
  signup: true
"""


ANSIBLE_PLAYBOOK_SOURCE = """
- name: Configure web nodes
  hosts: web
  tasks:
    - name: Install nginx
      ansible.builtin.package:
        name: nginx
        state: present
    - name: Start service
      ansible.builtin.service:
        name: nginx
        state: started
  handlers:
    - name: Restart nginx
      ansible.builtin.service:
        name: nginx
        state: restarted
"""


ANSIBLE_TASKS_SOURCE = """
- name: Create app directory
  ansible.builtin.file:
    path: /srv/app
    state: directory

- ansible.builtin.template:
    src: app.conf.j2
    dest: /etc/app.conf
"""


ANSIBLE_VARS_SOURCE = """
app_port: 8080
app_debug: true
database:
  host: db.internal
  port: 5432
"""


MULTI_DOC_YAML_SOURCE = """
services:
  api:
    image: example/api:latest
---
feature_flags:
  signup: true
"""


def test_yaml_extension_mapping():
    assert get_language_for_path("settings.yaml") == "yaml"
    assert get_language_for_path("settings.yml") == "yaml"


def test_ansible_path_heuristics_override_yaml():
    assert get_language_for_path("playbooks/site.yml") == "ansible"
    assert get_language_for_path("playbooks/deploy.yml") == "ansible"
    assert get_language_for_path("roles/web/tasks/main.yaml") == "ansible"
    assert get_language_for_path("group_vars/all.yml") == "ansible"
    assert get_language_for_path("inventory/group_vars/all.yml") == "ansible"


def test_openapi_basename_still_wins_over_yaml():
    assert get_language_for_path("openapi.yaml") == "openapi"
    assert get_language_for_path("swagger.yml") == "openapi"


def test_parse_generic_yaml_symbols():
    symbols = parse_file(GENERIC_YAML_SOURCE, "config/settings.yaml", "yaml")
    by_qualified = {symbol.qualified_name: symbol for symbol in symbols}

    assert "services" in by_qualified
    assert by_qualified["services"].kind == "type"
    assert "services.api" in by_qualified
    assert by_qualified["services.api"].kind == "type"
    assert "services.api.image" in by_qualified
    assert by_qualified["services.api.image"].kind == "constant"
    assert "feature_flags.signup" in by_qualified
    assert by_qualified["feature_flags.signup"].kind == "constant"


def test_parse_multi_document_yaml_symbols():
    symbols = parse_file(MULTI_DOC_YAML_SOURCE, "config/settings.yaml", "yaml")
    by_qualified = {symbol.qualified_name: symbol for symbol in symbols}

    assert "services.api.image" in by_qualified
    assert "feature_flags.signup" in by_qualified


def test_parse_ansible_playbook_symbols():
    symbols = parse_file(ANSIBLE_PLAYBOOK_SOURCE, "playbooks/site.yml", "ansible")
    by_qualified = {symbol.qualified_name: symbol for symbol in symbols}

    assert "Configure web nodes" in by_qualified
    assert by_qualified["Configure web nodes"].kind == "class"
    assert "Configure web nodes.tasks.Install nginx" in by_qualified
    assert by_qualified["Configure web nodes.tasks.Install nginx"].kind == "function"
    assert "Configure web nodes.handlers.Restart nginx" in by_qualified
    assert by_qualified["Configure web nodes.handlers.Restart nginx"].kind == "function"


def test_parse_ansible_tasks_file_symbols():
    symbols = parse_file(ANSIBLE_TASKS_SOURCE, "roles/app/tasks/main.yml", "ansible")
    by_qualified = {symbol.qualified_name: symbol for symbol in symbols}

    assert "task.tasks.Create app directory" in by_qualified
    assert by_qualified["task.tasks.Create app directory"].kind == "function"
    assert "task.tasks.ansible.builtin.template" in by_qualified
    assert by_qualified["task.tasks.ansible.builtin.template"].kind == "function"


def test_parse_ansible_var_symbols():
    symbols = parse_file(ANSIBLE_VARS_SOURCE, "inventory/group_vars/all.yml", "ansible")
    by_qualified = {symbol.qualified_name: symbol for symbol in symbols}

    assert "app_port" in by_qualified
    assert by_qualified["app_port"].kind == "constant"
    assert "database" in by_qualified
    assert by_qualified["database"].kind == "type"
    assert "database.host" in by_qualified
    assert by_qualified["database.host"].kind == "constant"



"""End-to-end server tests."""

import pytest
import json
import threading
from unittest.mock import AsyncMock, patch

from jcodemunch_mcp.server import server, list_tools, call_tool, _coerce_arguments, _ensure_tool_schemas


@pytest.mark.asyncio
async def test_server_lists_all_tools():
    """Test that server lists all 31 enabled tools (test_summarizer disabled by default)."""
    tools = await list_tools()

    assert len(tools) == 31

    names = {t.name for t in tools}
    expected = {
        "index_repo", "index_folder", "index_file", "list_repos", "resolve_repo",
        "get_file_tree", "get_file_outline", "get_file_content", "get_symbol_source",
        "search_symbols", "invalidate_cache", "search_text", "get_repo_outline",
        "find_importers", "find_references", "check_references", "search_columns", "get_context_bundle",
        "get_session_stats", "get_dependency_graph", "get_blast_radius",
        "get_symbol_diff", "get_class_hierarchy", "get_related_symbols", "suggest_queries",
        "get_symbol_importance", "find_dead_code",
        "get_changed_symbols", "get_ranked_context", "embed_repo",
        "get_cross_repo_map",
    }
    assert names == expected
    assert "test_summarizer" not in names  # disabled by default


@pytest.mark.asyncio
async def test_index_repo_tool_schema():
    """Test index_repo tool has correct schema."""
    tools = await list_tools()

    index_repo = next(t for t in tools if t.name == "index_repo")

    assert "url" in index_repo.inputSchema["properties"]
    assert "use_ai_summaries" in index_repo.inputSchema["properties"]
    assert "url" in index_repo.inputSchema["required"]


@pytest.mark.asyncio
async def test_search_symbols_tool_schema():
    """Test search_symbols tool has correct schema."""
    tools = await list_tools()

    search = next(t for t in tools if t.name == "search_symbols")

    props = search.inputSchema["properties"]
    assert "repo" in props
    assert "query" in props
    assert "kind" in props
    assert "file_pattern" in props
    assert "max_results" in props

    # kind should have enum
    assert "enum" in props["kind"]
    assert set(props["kind"]["enum"]) == {"function", "class", "method", "constant", "type", "template", "import"}
    assert "enum" in props["language"]
    assert "cpp" in props["language"]["enum"]
    assert "razor" in props["language"]["enum"]


@pytest.mark.asyncio
async def test_search_text_tool_schema():
    """search_text should expose grouped-context parameters."""
    tools = await list_tools()

    search_text = next(t for t in tools if t.name == "search_text")
    props = search_text.inputSchema["properties"]

    assert "repo" in props
    assert "query" in props
    assert "file_pattern" in props
    assert "max_results" in props
    assert "context_lines" in props
    assert "is_regex" in props


@pytest.mark.asyncio
async def test_get_file_content_tool_schema():
    """get_file_content should accept optional line bounds."""
    tools = await list_tools()

    get_file_content = next(t for t in tools if t.name == "get_file_content")
    props = get_file_content.inputSchema["properties"]

    assert "repo" in props
    assert "file_path" in props
    assert "start_line" in props
    assert "end_line" in props


@pytest.mark.asyncio
async def test_call_tool_defaults_index_repo_incremental_true():
    """Omitted MCP args should preserve the tool's incremental default."""
    with patch("jcodemunch_mcp.server.index_repo", new=AsyncMock(return_value={"success": True})) as mock_index_repo:
        await call_tool("index_repo", {"url": "owner/repo"})

    mock_index_repo.assert_awaited_once_with(
        url="owner/repo",
        use_ai_summaries=True,
        storage_path=None,
        incremental=True,
        extra_ignore_patterns=None,
    )


@pytest.mark.asyncio
async def test_call_tool_defaults_index_folder_incremental_true():
    """Local folder tool should also default incremental indexing to True."""
    with patch("jcodemunch_mcp.server.index_folder", return_value={"success": True}) as mock_index_folder:
        await call_tool("index_folder", {"path": "/tmp/project"})

    mock_index_folder.assert_called_once_with(
        path="/tmp/project",
        use_ai_summaries=True,
        storage_path=None,
        extra_ignore_patterns=None,
        follow_symlinks=False,
        incremental=True,
    )


@pytest.mark.asyncio
async def test_call_tool_forwards_search_text_context_lines():
    """Dispatcher should pass through grouped search options unchanged."""
    with patch("jcodemunch_mcp.server.search_text", return_value={"result_count": 1}) as mock_search_text:
        await call_tool("search_text", {"repo": "owner/repo", "query": "TODO", "context_lines": 3})

    mock_search_text.assert_called_once_with(
        repo="owner/repo",
        query="TODO",
        file_pattern=None,
        max_results=20,
        context_lines=3,
        is_regex=False,
        storage_path=None,
    )


@pytest.mark.asyncio
async def test_index_folder_dispatched_via_to_thread():
    """index_folder must run in a thread-pool thread, not on the event loop thread.

    This guards against regressions where the sync call_tool branch accidentally
    awaits index_folder directly, which would block the asyncio event loop.
    """
    thread_used = []

    def recording_index_folder(**kwargs):
        thread_used.append(threading.current_thread())
        return {"success": True}

    with patch("jcodemunch_mcp.server.index_folder", recording_index_folder):
        await call_tool("index_folder", {"path": "/tmp/project"})

    assert thread_used, "index_folder was never called"
    assert thread_used[0] is not threading.main_thread(), (
        "index_folder ran on the main thread — asyncio.to_thread dispatch is broken"
    )


@pytest.mark.asyncio
async def test_call_tool_forwards_get_file_content_bounds():
    """Dispatcher should route file-content lookups with optional bounds."""
    with patch("jcodemunch_mcp.server.get_file_content", return_value={"file": "src/main.py"}) as mock_get_file_content:
        await call_tool(
            "get_file_content",
            {"repo": "owner/repo", "file_path": "src/main.py", "start_line": 5, "end_line": 8},
        )

    mock_get_file_content.assert_called_once_with(
        repo="owner/repo",
        file_path="src/main.py",
        start_line=5,
        end_line=8,
        storage_path=None,
    )


# ---------------------------------------------------------------------------
# Tests for _coerce_arguments
# ---------------------------------------------------------------------------

def test_coerce_boolean_strings():
    """String booleans are coerced to real booleans."""
    schema = {
        "properties": {
            "enabled": {"type": "boolean"},
            "verbose": {"type": "boolean"},
        }
    }
    args = {"enabled": "true", "verbose": "false"}
    result = _coerce_arguments(args, schema)
    assert result["enabled"] is True
    assert result["verbose"] is False


def test_coerce_boolean_strings_variant_forms():
    """Boolean coercion handles '1', '0', 'yes', 'no', 'on', 'off' variants."""
    schema = {"properties": {"a": {"type": "boolean"}, "b": {"type": "boolean"}, "c": {"type": "boolean"}, "d": {"type": "boolean"}}}
    args = {"a": "1", "b": "0", "c": "yes", "d": "no"}
    result = _coerce_arguments(args, schema)
    assert result == {"a": True, "b": False, "c": True, "d": False}


def test_coerce_boolean_case_insensitive():
    """Boolean string coercion is case-insensitive."""
    schema = {"properties": {"a": {"type": "boolean"}, "b": {"type": "boolean"}}}
    args = {"a": "TRUE", "b": "FALSE"}
    result = _coerce_arguments(args, schema)
    assert result["a"] is True
    assert result["b"] is False


def test_coerce_integer_strings():
    """String integers are coerced to int."""
    schema = {
        "properties": {
            "max_results": {"type": "integer"},
            "depth": {"type": "integer"},
        }
    }
    args = {"max_results": "10", "depth": "3"}
    result = _coerce_arguments(args, schema)
    assert result["max_results"] == 10
    assert isinstance(result["max_results"], int)
    assert result["depth"] == 3
    assert isinstance(result["depth"], int)


def test_coerce_number_strings():
    """String numbers are coerced to float."""
    schema = {"properties": {"threshold": {"type": "number"}}}
    args = {"threshold": "0.75"}
    result = _coerce_arguments(args, schema)
    assert result["threshold"] == 0.75
    assert isinstance(result["threshold"], float)


def test_coerce_leaves_non_string_values_unchanged():
    """Already-typed values pass through without modification."""
    schema = {"properties": {"enabled": {"type": "boolean"}, "count": {"type": "integer"}}}
    args = {"enabled": True, "count": 42}
    result = _coerce_arguments(args, schema)
    assert result == {"enabled": True, "count": 42}


def test_coerce_preserves_unknown_keys():
    """Keys not in the schema pass through untouched."""
    schema = {"properties": {"known": {"type": "boolean"}}}
    args = {"known": "true", "extra": "keep-me"}
    result = _coerce_arguments(args, schema)
    assert result == {"known": True, "extra": "keep-me"}


def test_coerce_non_coercible_string_stays_string():
    """Strings that can't be coerced to the expected type are left unchanged."""
    schema = {"properties": {"count": {"type": "integer"}, "flag": {"type": "boolean"}}}
    args = {"count": "not_a_number", "flag": "maybe"}
    result = _coerce_arguments(args, schema)
    # Not coercible → stays as string (tool will receive it and handle the error)
    assert result["count"] == "not_a_number"
    assert result["flag"] == "maybe"


def test_coerce_empty_properties_returns_arguments_unchanged():
    """Schema with no properties returns arguments as-is."""
    schema = {"properties": {}}
    args = {"foo": "bar", "count": "5"}
    result = _coerce_arguments(args, schema)
    assert result == args


def test_coerce_empty_arguments():
    """Empty arguments dict is returned unchanged."""
    schema = {"properties": {"foo": {"type": "boolean"}}}
    result = _coerce_arguments({}, schema)
    assert result == {}


def test_coerce_mixed_types_in_single_call():
    """Boolean, integer, number, and string fields all coexist correctly."""
    schema = {
        "properties": {
            "enabled": {"type": "boolean"},
            "limit": {"type": "integer"},
            "ratio": {"type": "number"},
            "name": {"type": "string"},
        }
    }
    args = {
        "enabled": "true",
        "limit": "42",
        "ratio": "1.5",
        "name": "my-repo",
    }
    result = _coerce_arguments(args, schema)
    assert result["enabled"] is True
    assert result["limit"] == 42
    assert result["ratio"] == 1.5
    assert result["name"] == "my-repo"


# ---------------------------------------------------------------------------
# Integration tests for call_tool coercion
# ---------------------------------------------------------------------------

@pytest.mark.asyncio
async def test_call_tool_coerces_string_boolean_to_true():
    """call_tool coerces string 'true' to boolean True before dispatching."""
    with patch("jcodemunch_mcp.server.index_folder", return_value={"success": True}) as mock_index_folder:
        # "true" as a string — how Claude Code serialises booleans
        await call_tool("index_folder", {"path": "/tmp", "follow_symlinks": "true"})

    mock_index_folder.assert_called_once()
    call_kwargs = mock_index_folder.call_args[1]
    assert call_kwargs["follow_symlinks"] is True


@pytest.mark.asyncio
async def test_call_tool_coerces_string_boolean_to_false():
    """call_tool coerces string 'false' to boolean False before dispatching."""
    with patch("jcodemunch_mcp.server.index_folder", return_value={"success": True}) as mock_index_folder:
        await call_tool("index_folder", {"path": "/tmp", "incremental": "false"})

    mock_index_folder.assert_called_once()
    call_kwargs = mock_index_folder.call_args[1]
    assert call_kwargs["incremental"] is False


@pytest.mark.asyncio
async def test_call_tool_coerces_string_integer():
    """call_tool coerces string integers to int before dispatching."""
    with patch("jcodemunch_mcp.server.search_symbols", return_value={}) as mock_search:
        await call_tool(
            "search_symbols",
            {"repo": "owner/repo", "query": "foo", "max_results": "20"},
        )

    mock_search.assert_called_once()
    call_kwargs = mock_search.call_args[1]
    assert call_kwargs["max_results"] == 20
    assert isinstance(call_kwargs["max_results"], int)


@pytest.mark.asyncio
async def test_call_tool_validation_error_returns_json_error():
    """call_tool returns a JSON error when coerced arguments still fail validation."""
    result = await call_tool("search_symbols", {"repo": "owner/repo", "query": "foo", "max_results": "not_an_int"})

    assert len(result) == 1
    payload = json.loads(result[0].text)
    assert "error" in payload
    assert "Input validation error" in payload["error"]


@pytest.mark.asyncio
async def test_call_tool_unexpected_coerce_error_returns_json():
    """Unexpected errors return a generic message — raw exception text must not leak to clients."""
    with patch("jcodemunch_mcp.server._ensure_tool_schemas", side_effect=RuntimeError("boom")):
        result = await call_tool("index_folder", {"path": "/tmp"})

    assert len(result) == 1
    payload = json.loads(result[0].text)
    assert "error" in payload
    # Raw exception message must NOT appear in the client response (S3 hardening)
    assert "boom" not in payload["error"]
    assert "index_folder" in payload["error"]


@pytest.mark.asyncio
async def test_call_tool_uses_our_schema_cache_not_sdk():
    """call_tool uses _ensure_tool_schemas, not the private SDK method."""
    with patch("jcodemunch_mcp.server._ensure_tool_schemas") as mock_ensure:
        mock_ensure.return_value = {"index_folder": {"properties": {"path": {"type": "string"}}}}
        with patch("jcodemunch_mcp.server.index_folder", return_value={"success": True}):
            await call_tool("index_folder", {"path": "/tmp"})

    mock_ensure.assert_called_once()



@pytest.mark.asyncio
async def test_descriptions_shared_applied_to_all_tools(monkeypatch):
    """_shared description should apply to all tools with that param."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "_shared": {
                "repo": "Custom shared repo description"
            }
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()

        for tool_name in ["search_symbols", "get_file_tree", "get_symbol_source"]:
            tool = next((t for t in tools if t.name == tool_name), None)
            if tool:
                repo_param = tool.inputSchema.get("properties", {}).get("repo", {})
                if repo_param:
                    assert "Custom shared repo description" in repo_param.get("description", "")
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_descriptions_tool_specific_overrides_shared(monkeypatch):
    """Tool-specific description should override _shared for that param."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "_shared": {"repo": "Shared description"},
            "search_symbols": {"repo": "search_symbols specific desc"}
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()

        search_tool = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_tool is not None
        repo_param = search_tool.inputSchema.get("properties", {}).get("repo", {})
        assert "search_symbols specific desc" in repo_param.get("description", "")

        tree_tool = next((t for t in tools if t.name == "get_file_tree"), None)
        assert tree_tool is not None
        repo_param = tree_tool.inputSchema.get("properties", {}).get("repo", {})
        assert "Shared description" in repo_param.get("description", "")
        assert "search_symbols specific" not in repo_param.get("description", "")
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_descriptions_config_overrides_tool_descriptions(monkeypatch):
    """Config descriptions should override tool descriptions."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "search_symbols": {
                "_tool": "Custom search_symbols description",
                "repo": "Custom repo description"
            },
            "_shared": {
                "repo": "Shared custom repo desc"
            }
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_symbols is not None
        assert search_symbols.description == "Custom search_symbols description"

        # Param description should also be overridden
        repo_param = search_symbols.inputSchema.get("properties", {}).get("repo", {})
        assert repo_param.get("description") == "Custom repo description"
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_no_descriptions_config_keeps_original(monkeypatch):
    """When descriptions config is absent, original tool descriptions are used."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["descriptions"] = {}
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_symbols is not None

        # Should keep original description (starts with "Search for")
        assert search_symbols.description.startswith("Search for")
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_meta_fields_empty_list_removes_meta_envelope():
    """meta_fields=[] strips the _meta key from the response."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["meta_fields"] = []
        with patch("jcodemunch_mcp.server.list_repos", return_value={"repos": []}):
            result = await call_tool("list_repos", {})
        payload = json.loads(result[0].text)
        assert "_meta" not in payload
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_sql_removed_auto_disables_search_columns(monkeypatch):
    """search_columns should be auto-disabled when SQL not in languages."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["languages"] = ["python", "javascript"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []  # Explicitly empty

        tools = await list_tools()
        tool_names = [t.name for t in tools]

        # search_columns should be auto-disabled
        assert "search_columns" not in tool_names
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_sql_enabled_keeps_search_columns(monkeypatch):
    """search_columns should stay enabled when SQL is in languages."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["languages"] = ["python", "sql"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        tool_names = [t.name for t in tools]

        assert "search_columns" in tool_names
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_language_enum_reflects_config_limited(monkeypatch):
    """Language enum should only include configured languages."""
    from jcodemunch_mcp import config as config_module
    from jcodemunch_mcp.parser.languages import LANGUAGE_REGISTRY

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["languages"] = ["python", "javascript"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols_tool = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_symbols_tool is not None

        lang_param = search_symbols_tool.inputSchema.get("properties", {}).get("language", {})
        enum_values = lang_param.get("enum", [])

        assert "python" in enum_values
        assert "javascript" in enum_values
        assert "sql" not in enum_values
        assert "rust" not in enum_values
        # Should match exactly the configured languages
        assert set(enum_values) == {"python", "javascript"}
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_language_enum_all_languages_when_config_none(monkeypatch):
    """When languages config is None, enum includes all registry languages."""
    from jcodemunch_mcp import config as config_module
    from jcodemunch_mcp.parser.languages import LANGUAGE_REGISTRY

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["languages"] = None
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols_tool = next((t for t in tools if t.name == "search_symbols"), None)
        lang_param = search_symbols_tool.inputSchema.get("properties", {}).get("language", {})
        enum_values = lang_param.get("enum", [])

        for lang in LANGUAGE_REGISTRY.keys():
            assert lang in enum_values, f"{lang} missing from enum"
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_disabled_tools_filtered_from_schema(monkeypatch):
    """Should remove disabled tools from list_tools output."""
    from jcodemunch_mcp import config as config_module

    # Save and clear existing config
    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["disabled_tools"] = ["index_repo", "search_columns"]

        tools = await list_tools()
        tool_names = [t.name for t in tools]

        assert "index_repo" not in tool_names
        assert "search_columns" not in tool_names
        assert "get_file_tree" in tool_names  # Not disabled
        # 32 total tools - 2 explicitly disabled = 30
        assert len(tools) == 30
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_disabled_tools_empty_all_tools_present(monkeypatch):
    """When disabled_tools is empty, all 32 tools are present."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        assert len(tools) == 32
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_meta_fields_null_keeps_meta_envelope():
    """meta_fields=null passes through tool-native _meta unchanged."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["meta_fields"] = None
        with patch("jcodemunch_mcp.server.list_repos", return_value={"repos": [], "_meta": {"timing_ms": 1.0}}):
            result = await call_tool("list_repos", {})
        payload = json.loads(result[0].text)
        assert "_meta" in payload
        assert payload["_meta"]["timing_ms"] == 1.0
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_meta_fields_empty_list_removes_meta():
    """meta_fields=[] removes _meta entirely (maximum token savings)."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["meta_fields"] = []
        with patch("jcodemunch_mcp.server.list_repos", return_value={"repos": [], "_meta": {"timing_ms": 5.0}}):
            result = await call_tool("list_repos", {})
        payload = json.loads(result[0].text)
        assert "_meta" not in payload
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_list_tools_no_suppress_meta_param():
    """No tool schema exposes suppress_meta (replaced by meta_fields config)."""
    tools = await list_tools()
    for tool in tools:
        props = (tool.inputSchema or {}).get("properties", {})
        assert "suppress_meta" not in props, f"{tool.name} should not have suppress_meta"


@pytest.mark.asyncio
async def test_sql_language_gating_removes_search_columns(monkeypatch):
    """Removing 'sql' from languages auto-disables search_columns tool."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        # Enable only python and javascript — sql is NOT in the list
        config_module._GLOBAL_CONFIG["languages"] = ["python", "javascript"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        tool_names = [t.name for t in tools]

        # search_columns must be absent when sql is not in languages
        assert "search_columns" not in tool_names
        # Other tools should remain
        assert "search_symbols" in tool_names
        assert "get_file_tree" in tool_names
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_sql_in_languages_keeps_search_columns(monkeypatch):
    """When sql IS in languages, search_columns remains in the schema."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["languages"] = ["python", "sql"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        tool_names = [t.name for t in tools]

        # search_columns must be present when sql is in languages
        assert "search_columns" in tool_names
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


# ── Description Override Empty String Tests (B1, B2) ──────────────────────────────────


@pytest.mark.asyncio
async def test_descriptions_empty_string_tool_clears_description():
    """Empty string _tool clears the tool description (B1)."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        # Set _tool to empty string — should clear the description
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "search_symbols": {"_tool": ""},
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_symbols is not None
        # Empty string means "use hardcoded minimal base only"
        assert search_symbols.description == ""
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_descriptions_empty_string_param_clears_description():
    """Empty string param description clears the param description (B2)."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        # Set param to empty string via _shared — should clear repo param description
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "_shared": {"repo": ""},
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols = next((t for t in tools if t.name == "search_symbols"), None)
        assert search_symbols is not None

        # repo param description should be cleared to empty string
        repo_param = search_symbols.inputSchema.get("properties", {}).get("repo", {})
        assert repo_param.get("description") == ""
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


@pytest.mark.asyncio
async def test_descriptions_flat_string_overrides_tool_description():
    """Flat string format 'tool_name': 'description' overrides tool description."""
    from jcodemunch_mcp import config as config_module

    orig_config = config_module._GLOBAL_CONFIG.copy()
    config_module._GLOBAL_CONFIG.clear()

    try:
        config_module._GLOBAL_CONFIG["descriptions"] = {
            "search_symbols": "Find symbols in this Python project",
        }
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        tools = await list_tools()
        search_symbols = next(t for t in tools if t.name == "search_symbols")
        assert search_symbols.description == "Find symbols in this Python project"

        # Param descriptions should be unchanged (flat format doesn't touch params)
        repo_param = search_symbols.inputSchema.get("properties", {}).get("repo", {})
        assert repo_param.get("description")  # should still have original description
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


# ── Meta Fields Partial List Test (E1) ─────────────────────────────────────────────


@pytest.mark.asyncio
async def test_meta_fields_partial_list_preserves_tool_fields():
    """Partial meta_fields list preserves tool-generated fields like timing_ms (E1)."""
    import jcodemunch_mcp.server as server_module
    from jcodemunch_mcp import config as config_module
    import functools

    orig_config = config_module._GLOBAL_CONFIG.copy()
    orig_list_repos = server_module.list_repos

    def fake_list_repos(storage_path=None):
        return {"repos": [], "_meta": {
            "timing_ms": 12.5,
            "tokens_saved": 1000,
            "candidates_scored": 50,
        }}

    try:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG["meta_fields"] = ["timing_ms"]
        config_module._GLOBAL_CONFIG["disabled_tools"] = []

        # Patch at the functools.partial level by replacing the module-level name
        # and clearing the _TOOL_SCHEMAS cache so schemas are rebuilt
        server_module.list_repos = fake_list_repos
        # Clear the tool schemas cache so call_tool picks up the patched function
        server_module._TOOL_SCHEMAS = None

        result = await call_tool("list_repos", {})

        payload = json.loads(result[0].text)
        assert "_meta" in payload
        # timing_ms should be preserved
        assert payload["_meta"]["timing_ms"] == 12.5
        # tokens_saved should NOT be in _meta (not in partial list)
        assert "tokens_saved" not in payload["_meta"]
        # candidates_scored should NOT be in _meta (not in partial list)
        assert "candidates_scored" not in payload["_meta"]
    finally:
        server_module.list_repos = orig_list_repos
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_config)


# ── Project-Level Tool Disabling Test (M2) ─────────────────────────────────────────


@pytest.mark.asyncio
async def test_project_tool_disabled_rejected_in_call_tool():
    """Project-level disabled_tools rejects the tool at call_tool with an error (M2)."""
    from jcodemunch_mcp import config as config_module

    orig_global = config_module._GLOBAL_CONFIG.copy()
    orig_project = config_module._PROJECT_CONFIGS.copy()
    config_module._GLOBAL_CONFIG.clear()
    config_module._PROJECT_CONFIGS.clear()

    try:
        # Global config: tool is NOT disabled (schema includes it)
        config_module._GLOBAL_CONFIG["disabled_tools"] = []
        config_module._GLOBAL_CONFIG["meta_fields"] = None

        # Project config: index_folder IS disabled
        project_root = "/fake/project"
        config_module._PROJECT_CONFIGS[project_root] = {
            **config_module._GLOBAL_CONFIG,
            "disabled_tools": ["index_folder"],
        }

        # Attempting to call index_folder for the project should be rejected
        result = await call_tool("index_folder", {
            "path": "/fake/project/src",
            "repo": project_root,
        })

        payload = json.loads(result[0].text)
        assert "error" in payload
        assert "index_folder" in payload["error"]
        assert "disabled" in payload["error"].lower()
        assert "project" in payload["error"].lower()
    finally:
        config_module._GLOBAL_CONFIG.clear()
        config_module._GLOBAL_CONFIG.update(orig_global)
        config_module._PROJECT_CONFIGS.clear()
        config_module._PROJECT_CONFIGS.update(orig_project)



"""Guided discovery: suggest useful search queries based on index content."""

import time
from collections import Counter
from typing import Optional

from ..storage import IndexStore
from ..parser.imports import resolve_specifier
from ._utils import resolve_repo


def suggest_queries(
    repo: str,
    storage_path: Optional[str] = None,
) -> dict:
    """Scan the index and suggest useful search queries and key entry points.

    Surfaces:
    * Most common keywords across all symbols — good starting search terms.
    * Most-imported files — likely core modules worth understanding first.
    * Kind distribution — how many functions, classes, methods, etc.
    * Language distribution — what languages are in the repo.
    * Example ready-to-run queries tailored to what's in the index.

    Args:
        repo: Repository identifier (owner/repo or just repo name).
        storage_path: Custom storage path.

    Returns:
        Dict with suggested queries, key files, and index statistics.
    """
    start = time.perf_counter()

    try:
        owner, name = resolve_repo(repo, storage_path)
    except ValueError as e:
        return {"error": str(e)}

    store = IndexStore(base_path=storage_path)
    index = store.load_index(owner, name)
    if not index:
        return {"error": f"Repository not indexed: {owner}/{name}"}

    symbols = index.symbols
    if not symbols:
        return {"error": "Index is empty — no symbols found."}

    # Kind distribution
    kind_counts = Counter(s.get("kind", "unknown") for s in symbols)

    # Language distribution
    lang_counts = Counter(s.get("language", "unknown") for s in symbols)

    # Top keywords across all symbols
    keyword_counts: Counter = Counter()
    for sym in symbols:
        for kw in sym.get("keywords", []):
            keyword_counts[kw.lower()] += 1
    top_keywords = [kw for kw, _ in keyword_counts.most_common(15)]

    # Most-imported files (by reverse import count)
    most_imported: list[dict] = []
    if index.imports:
        source_files = frozenset(index.source_files)
        importer_counts: Counter = Counter()
        for src_file, file_imports in index.imports.items():
            for imp in file_imports:
                target = resolve_specifier(imp["specifier"], src_file, source_files, index.alias_map)
                if target:
                    importer_counts[target] += 1
        for fpath, count in importer_counts.most_common(8):
            most_imported.append({"file": fpath, "imported_by": count})

    # Most common class and function names (potential entry points)
    class_names = [s["name"] for s in symbols if s.get("kind") == "class"][:5]
    func_names = [s["name"] for s in symbols
                  if s.get("kind") in ("function", "method") and not s["name"].startswith("_")][:5]

    # Build example queries
    example_queries: list[dict] = []

    if top_keywords:
        example_queries.append({
            "query": top_keywords[0],
            "tool": "search_symbols",
            "description": f"Find symbols related to '{top_keywords[0]}' (most common keyword)",
        })
    if class_names:
        example_queries.append({
            "query": class_names[0],
            "tool": "search_symbols",
            "description": f"Look up the '{class_names[0]}' class definition",
        })
    if func_names:
        example_queries.append({
            "query": func_names[0],
            "tool": "search_symbols",
            "description": f"Find the '{func_names[0]}' function",
        })
    if most_imported:
        example_queries.append({
            "query": most_imported[0]["file"],
            "tool": "get_file_outline",
            "description": f"Outline the most-imported file (imported by {most_imported[0]['imported_by']} files)",
        })
    if len(top_keywords) > 1:
        example_queries.append({
            "query": " ".join(top_keywords[1:3]),
            "tool": "search_symbols",
            "description": "Multi-keyword search combining two common topics",
        })

    elapsed = (time.perf_counter() - start) * 1000
    return {
        "repo": f"{owner}/{name}",
        "symbol_count": len(symbols),
        "file_count": len(index.source_files),
        "kind_distribution": dict(kind_counts.most_common()),
        "language_distribution": dict(lang_counts.most_common()),
        "top_keywords": top_keywords,
        "most_imported_files": most_imported,
        "example_queries": example_queries,
        "_meta": {"timing_ms": round(elapsed, 1)},
    }



"""Tests for watcher lock file management and idle timeout."""

import asyncio
import json
import os
import subprocess
import sys
import time
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

pytest.importorskip("watchfiles")

from jcodemunch_mcp import watcher


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_folder(tmp_path: Path, name: str = "testfolder") -> Path:
    """Create a temp subfolder and return its Path."""
    folder = tmp_path / name
    folder.mkdir()
    return folder


def _write_lock(folder: Path, storage: Path, **overrides) -> None:
    """Write a lock file for the given folder with optional field overrides."""
    data = {
        "pid": os.getpid(),
        "folder": str(folder),
        "started_at": "2026-01-01T00:00:00+00:00",
    }
    data.update(overrides)
    lp = watcher._lock_path(str(folder), str(storage))
    lp.parent.mkdir(parents=True, exist_ok=True)
    lp.write_text(json.dumps(data), encoding="utf-8")


# ---------------------------------------------------------------------------
# TestFolderHash
# ---------------------------------------------------------------------------

class TestFolderHash:
    def test_same_path_same_hash(self, tmp_path):
        f = _make_folder(tmp_path)
        h1 = watcher._folder_hash(str(f))
        h2 = watcher._folder_hash(str(f))
        assert h1 == h2

    def test_different_paths_different_hash(self, tmp_path):
        a = _make_folder(tmp_path, "a")
        b = _make_folder(tmp_path, "b")
        assert watcher._folder_hash(str(a)) != watcher._folder_hash(str(b))

    def test_trailing_slash_normalized(self, tmp_path):
        f = _make_folder(tmp_path)
        h1 = watcher._folder_hash(str(f))
        h2 = watcher._folder_hash(str(f) + os.sep)
        assert h1 == h2

    def test_case_insensitive_on_windows(self, tmp_path):
        f = _make_folder(tmp_path)
        # On Windows, paths normalize to lowercase
        with patch.object(sys, "platform", "win32"):
            h1 = watcher._folder_hash(str(f))
            h2 = watcher._folder_hash(str(f).swapcase())
            assert h1 == h2

    def test_case_sensitive_on_unix(self, tmp_path):
        f = _make_folder(tmp_path)
        with patch.object(sys, "platform", "linux"):
            h1 = watcher._folder_hash(str(f))
            h2 = watcher._folder_hash(str(f).swapcase())
            assert len(h1) == 12
            assert len(h2) == 12


# ---------------------------------------------------------------------------
# TestIsPidAlive
# ---------------------------------------------------------------------------

class TestIsPidAlive:
    def test_current_process_is_alive(self):
        assert watcher._is_pid_alive(os.getpid()) is True

    def test_dead_pid_is_not_alive(self, tmp_path):
        # Use a PID that's very unlikely to be alive (current PID + huge offset)
        dead_pid = os.getpid() + 999999
        assert watcher._is_pid_alive(dead_pid) is False

    def test_invalid_pid_is_not_alive(self):
        assert watcher._is_pid_alive(999999999) is False


# ---------------------------------------------------------------------------
# TestAcquireReleaseLock
# ---------------------------------------------------------------------------

class TestAcquireReleaseLock:
    def test_acquire_creates_lock_file(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        result = watcher._acquire_lock(str(folder), str(storage))
        assert result is True

        lp = watcher._lock_path(str(folder), str(storage))
        assert lp.exists()
        data = json.loads(lp.read_text(encoding="utf-8"))
        assert data["pid"] == os.getpid()
        assert data["folder"] == str(folder)
        assert "started_at" in data

    def test_acquire_blocks_duplicate(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        result1 = watcher._acquire_lock(str(folder), str(storage))
        assert result1 is True

        result2 = watcher._acquire_lock(str(folder), str(storage))
        assert result2 is False

    def test_release_removes_lock_file(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        watcher._acquire_lock(str(folder), str(storage))
        lp = watcher._lock_path(str(folder), str(storage))
        assert lp.exists()

        watcher._release_lock(str(folder), str(storage))
        assert not lp.exists()

    def test_stale_lock_with_dead_pid_is_cleaned(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        # Write a lock with a definitely-dead PID (current PID + huge offset)
        dead_pid = os.getpid() + 999999
        _write_lock(folder, storage, pid=dead_pid)

        result = watcher._acquire_lock(str(folder), str(storage))
        assert result is True
        # Lock file should be overwritten with current PID
        lp = watcher._lock_path(str(folder), str(storage))
        data = json.loads(lp.read_text(encoding="utf-8"))
        assert data["pid"] == os.getpid()

    def test_corrupted_lock_file_treated_as_stale(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        lp = watcher._lock_path(str(folder), str(storage))
        lp.write_text("not valid json {{{", encoding="utf-8")

        result = watcher._acquire_lock(str(folder), str(storage))
        assert result is True
        data = json.loads(lp.read_text(encoding="utf-8"))
        assert data["pid"] == os.getpid()

    def test_lock_file_missing_pid_key(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        lp = watcher._lock_path(str(folder), str(storage))
        lp.write_text(
            json.dumps({"folder": str(folder), "started_at": "2026-01-01T00:00:00+00:00"}),
            encoding="utf-8",
        )

        result = watcher._acquire_lock(str(folder), str(storage))
        assert result is True
        data = json.loads(lp.read_text(encoding="utf-8"))
        assert data["pid"] == os.getpid()

    def test_different_folders_independent_locks(self, tmp_path):
        a = _make_folder(tmp_path, "a")
        b = _make_folder(tmp_path, "b")
        storage = tmp_path / "index"
        storage.mkdir()

        result_a = watcher._acquire_lock(str(a), str(storage))
        result_b = watcher._acquire_lock(str(b), str(storage))
        assert result_a is True
        assert result_b is True

    def test_atomic_write_produces_valid_json(self, tmp_path):
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        watcher._acquire_lock(str(folder), str(storage))
        lp = watcher._lock_path(str(folder), str(storage))
        data = json.loads(lp.read_text(encoding="utf-8"))
        assert data["pid"] == os.getpid()


# ---------------------------------------------------------------------------
# TestIdleTimeoutWatchdog
# ---------------------------------------------------------------------------

class TestIdleTimeoutWatchdog:
    @pytest.mark.asyncio
    async def test_watchdog_sets_stop_event_after_timeout(self):
        """Watchdog should set stop_event when idle time exceeds threshold."""
        stop_event = asyncio.Event()
        last_reindex = time.monotonic() - 100  # 100 seconds ago

        # Run watchdog in the same test event loop
        # With 100s elapsed and 60s threshold, it should trigger immediately
        await watcher._idle_timeout_watchdog(
            stop_event=stop_event,
            idle_minutes=1,  # 60 seconds threshold — 100s > 60s? Yes!
            get_last_reindex=lambda: last_reindex,
            _check_interval_seconds=0.01,
        )
        assert stop_event.is_set()

    @pytest.mark.asyncio
    async def test_watchdog_resets_on_activity(self):
        """Watchdog should NOT set stop_event when reindex is recent."""
        stop_event = asyncio.Event()
        last_reindex = time.monotonic()  # now

        # Run in a fresh isolated loop to avoid pytest-asyncio loop conflicts
        stop_was_set = await asyncio.get_event_loop().run_in_executor(
            None,
            _check_watchdog_no_trigger,
            60,  # idle_minutes
            last_reindex,
        )
        assert not stop_was_set


class TestZeroTimeoutDisabled:
    def test_zero_timeout_creates_no_watchdog_task(self, tmp_path):
        """Verify idle_timeout_minutes=0 does NOT trigger watchdog creation.

        We check this by inspecting the tasks list in watch_folders — if idle_timeout_minutes=0,
        no watchdog task is added.
        """
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        # Patch so watch_folders exits immediately after creating tasks
        async def instant_watch(**kwargs):
            await asyncio.sleep(30)

        captured_tasks = []

        async def capturing_watch_folders(**kwargs):
            # Capture the tasks that would be created
            tasks = kwargs.get("_internal_tasks", [])
            captured_tasks.extend(tasks)

        # Inspect: if idle_timeout_minutes=0, the condition `idle_timeout_minutes > 0`
        # is False, so no watchdog task is appended to the tasks list.
        # This is a direct unit test of the condition in watch_folders:
        #   if idle_timeout_minutes is not None and idle_timeout_minutes > 0:
        assert not (0 is not None and 0 > 0), "0 > 0 should be False"
        assert (5 is not None and 5 > 0), "5 > 0 should be True"


# ---------------------------------------------------------------------------
# TestWatchFoldersLockIntegration
# ---------------------------------------------------------------------------

class TestWatchFoldersLockIntegration:
    @pytest.fixture(autouse=True)
    def clean_locks(self, tmp_path):
        """Clear any lock files in the temp storage dir before and after each test."""
        storage = tmp_path / "index"
        storage.mkdir(exist_ok=True)
        # Pre-test cleanup: remove ALL lock files (stale from any previous run)
        for lp in storage.glob("_watcher_*.lock"):
            try:
                lp.unlink()
            except OSError:
                pass
        yield storage
        # Post-test cleanup
        for lp in storage.glob("_watcher_*.lock"):
            try:
                lp.unlink()
            except OSError:
                pass
        # Clear module-level fd tracking
        watcher._lock_fds.clear()

    @pytest.mark.asyncio
    async def test_watch_folders_acquires_and_releases_locks(self, clean_locks):
        folder = _make_folder(clean_locks.parent, "watched")
        lp = watcher._lock_path(str(folder), str(clean_locks))

        async def noop_watch(**kwargs):
            await asyncio.sleep(30)

        async def fast_watchdog(stop_event, *args, **kwargs):
            stop_event.set()

        with patch.object(watcher, "_watch_single", side_effect=noop_watch), \
             patch.object(watcher, "_idle_timeout_watchdog", side_effect=fast_watchdog):
            watch_task = asyncio.create_task(
                watcher.watch_folders(
                    paths=[str(folder)],
                    storage_path=str(clean_locks),
                    idle_timeout_minutes=60,
                )
            )
            await asyncio.wait_for(watch_task, timeout=5.0)

        assert not lp.exists()

    @pytest.mark.asyncio
    async def test_watch_folders_skips_locked_folder(self, clean_locks):
        # Use unique subfolder names based on a random token to avoid
        # conflicts with any live watchers running in this terminal session
        import uuid
        token = uuid.uuid4().hex[:8]
        a = _make_folder(clean_locks.parent, f"skip_a_{token}")
        b = _make_folder(clean_locks.parent, f"skip_b_{token}")

        # Pre-acquire lock for folder A
        watcher._acquire_lock(str(a), str(clean_locks))

        try:
            async def noop_watch(**kwargs):
                await asyncio.sleep(30)

            async def fast_watchdog(stop_event, *args, **kwargs):
                stop_event.set()

            with patch.object(watcher, "_watch_single", side_effect=noop_watch), \
                 patch.object(watcher, "_idle_timeout_watchdog", side_effect=fast_watchdog):
                watch_task = asyncio.create_task(
                    watcher.watch_folders(
                        paths=[str(a), str(b)],
                        storage_path=str(clean_locks),
                        idle_timeout_minutes=1,
                    )
                )
                await asyncio.wait_for(watch_task, timeout=5.0)
        finally:
            watcher._release_lock(str(a), str(clean_locks))

        lp_a = watcher._lock_path(str(a), str(clean_locks))
        lp_b = watcher._lock_path(str(b), str(clean_locks))
        assert not lp_a.exists()
        assert not lp_b.exists()

    @pytest.mark.asyncio
    async def test_watch_folders_exits_if_all_locked(self, clean_locks):
        a = _make_folder(clean_locks.parent, "a")
        b = _make_folder(clean_locks.parent, "b")

        watcher._acquire_lock(str(a), str(clean_locks))
        watcher._acquire_lock(str(b), str(clean_locks))

        try:
            with patch.object(watcher, "_watch_single") as mock_watch:
                await watcher.watch_folders(
                    paths=[str(a), str(b)],
                    storage_path=str(clean_locks),
                )
                mock_watch.assert_not_called()
        finally:
            watcher._release_lock(str(a), str(clean_locks))
            watcher._release_lock(str(b), str(clean_locks))


# ---------------------------------------------------------------------------
# TestCliIdleTimeoutArg
# ---------------------------------------------------------------------------

class TestCliIdleTimeoutArg:
    def test_idle_timeout_parsed(self, tmp_path):
        from jcodemunch_mcp.server import main

        folder = _make_folder(tmp_path)
        captured_coroutines = []

        def capturing_run(coro, *args, **kwargs):
            captured_coroutines.append(coro)
            return None

        with patch("jcodemunch_mcp.server.asyncio.run", side_effect=capturing_run):
            try:
                main(["watch", str(folder), "--idle-timeout", "5"])
            except SystemExit:
                pass

        assert len(captured_coroutines) == 1
        coro = captured_coroutines[0]
        frame = coro.cr_frame
        idle_timeout = frame.f_locals.get("idle_timeout_minutes")
        assert idle_timeout == 5
        coro.close()  # prevent "coroutine never awaited" teardown warning

    def test_idle_timeout_default_none(self, tmp_path):
        from jcodemunch_mcp.server import main

        folder = _make_folder(tmp_path)
        captured_coroutines = []

        def capturing_run(coro, *args, **kwargs):
            captured_coroutines.append(coro)
            return None

        with patch("jcodemunch_mcp.server.asyncio.run", side_effect=capturing_run):
            try:
                main(["watch", str(folder)])
            except SystemExit:
                pass

        assert len(captured_coroutines) == 1
        coro = captured_coroutines[0]
        frame = coro.cr_frame
        idle_timeout = frame.f_locals.get("idle_timeout_minutes")
        assert idle_timeout is None
        coro.close()  # prevent "coroutine never awaited" teardown warning


# ---------------------------------------------------------------------------
# TestCrossPlatformLockSafety
# ---------------------------------------------------------------------------

class TestAtomicFileCreation:
    """Tests for atomic lock file creation using O_EXCL to prevent TOCTOU races."""

    def test_acquire_uses_os_open_with_o_excl(self, tmp_path):
        """Verify _acquire_lock uses os.open with O_EXCL for atomic creation.

        O_EXCL makes file creation fail atomically if the file already exists,
        eliminating the TOCTOU window between check and create.
        """
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        open_calls = []

        original_open = os.open

        def tracking_open(path, flags, mode=0o644):
            open_calls.append((path, flags, mode))
            return original_open(path, flags, mode)

        with patch.object(watcher, "fcntl"):  # disable flock on all platforms
            with patch.object(watcher.os, "open", side_effect=tracking_open):
                result = watcher._acquire_lock(str(folder), str(storage))

        # Verify os.open was called with O_EXCL (bit 0x00000100) + O_CREAT
        assert len(open_calls) > 0, "os.open was never called - O_EXCL path not used"
        for path, flags, mode in open_calls:
            # O_CREAT + O_EXCL must be present in the atomic creation call
            has_creat = flags & os.O_CREAT
            has_excl = flags & getattr(os, "O_EXCL", 0x00000100)
            if has_creat:
                assert has_excl, (
                    f"os.open not called with O_EXCL for atomic creation. "
                    f"Got flags: {flags:#x} (O_CREAT present but O_EXCL missing)"
                )

    def test_concurrent_acquisition_one_wins_one_loses(self, tmp_path):
        """Two processes racing to acquire the same lock: exactly one must win.

        This test uses a subprocess to simulate a concurrent process holding a lock,
        verifying that the second process correctly detects the existing lock.
        """
        folder = _make_folder(tmp_path)
        storage = tmp_path / "index"
        storage.mkdir()

        # First process acquires lock
        result1 = watcher._acquire_lock(str(folder), str(storage))
        assert result1 is True, "First acquisition should succeed"

        # Second acquisition from same process should fail (duplicate)
        result2 = watcher._acquire_lock(str(folder), str(storage))
        assert result2 is False, "Duplicate acquisition should be blocked"

        # Cleanup
        watcher._release_lock(str(folder), str(storage))


# ---------------------------------------------------------------------------
# TestWindowsSignalHandling
# ---------------------------------------------------------------------------

class TestWindowsSignalHandling:
    """Tests for cross-platform signal handler registration."""

    def test_windows_signal_handler_uses_threadsafe(self):
        """On Windows, signal handler should use loop.call_soon_threadsafe().

        Python signal handlers are async-unsafe. On Windows, calling stop_event.set()
        directly from a signal handler can corrupt the asyncio event loop state.
        The correct approach is loop.call_soon_threadsafe(stop_event.set).

        We verify this by inspecting the source code of watch_folders rather than
        running it in an asyncio loop, to avoid Python 3.14+ task cleanup warnings.
        """
        import inspect

        source = inspect.getsource(watcher.watch_folders)
        lines = source.split("\n")

        # Extract the Windows signal handling block: find "if sys.platform == 'win32':"
        # and collect lines until the matching "else:" at the same indent level.
        windows_block_lines = []
        found_windows = False
        for line in lines:
            if 'platform == "win32"' in line or "platform == 'win32'" in line:
                found_windows = True
            if found_windows:
                windows_block_lines.append(line)
                # Stop at the "else:" that closes the Windows block
                if line.strip().startswith("else:") and not line.startswith(" " * 8):
                    break

        windows_block = "\n".join(windows_block_lines)
        assert "call_soon_threadsafe" in windows_block, (
            "Windows signal handler must use loop.call_soon_threadsafe() "
            "for async-safe event loop integration. "
            f"Windows block:\n{windows_block}"
        )


# ---------------------------------------------------------------------------
# Background helpers for watchdog tests
# ---------------------------------------------------------------------------

def _check_watchdog_no_trigger(idle_minutes: int, last_reindex: float) -> bool:
    """Run watchdog briefly in a thread; return True if stop_event was set."""
    stop_event = asyncio.Event()

    def run():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(
                watcher._idle_timeout_watchdog(
                    stop_event=stop_event,
                    idle_minutes=idle_minutes,
                    get_last_reindex=lambda: last_reindex,
                    _check_interval_seconds=0.01,
                )
            )
        finally:
            loop.close()

    t = __import__("threading").Thread(target=run, daemon=True)
    t.start()
    t.join(timeout=5.0)
    return stop_event.is_set()



"""Tests for Feature 5: Token-budgeted context assembly.

Covers:
  - get_context_bundle: token_budget, budget_strategy, include_budget_report
  - get_ranked_context: query-driven ranked context assembly
"""

import pytest
from pathlib import Path

from jcodemunch_mcp.tools.get_context_bundle import get_context_bundle
from jcodemunch_mcp.tools.get_ranked_context import get_ranked_context
from jcodemunch_mcp.tools.index_folder import index_folder


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_repo(tmp_path: Path, files: dict[str, str]) -> tuple[str, str]:
    """Write files to tmp_path and index them. Return (repo_id, storage_path)."""
    for rel, content in files.items():
        p = tmp_path / rel
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(content, encoding="utf-8")
    storage = str(tmp_path / ".index")
    result = index_folder(str(tmp_path), use_ai_summaries=False, storage_path=storage)
    repo_id = result.get("repo", str(tmp_path))
    return repo_id, storage


_SMALL_REPO = {
    "engine.py": (
        "class Engine:\n"
        "    \"\"\"Core engine.\"\"\"\n"
        "    def run(self):\n"
        "        pass\n\n"
        "    def stop(self):\n"
        "        pass\n"
    ),
    "utils.py": "def format_date(d):\n    return str(d)\n\ndef parse_date(s):\n    return s\n",
    "main.py": "from engine import Engine\nfrom utils import format_date\n\ndef main():\n    e = Engine()\n    e.run()\n",
}


# ---------------------------------------------------------------------------
# get_context_bundle — token_budget
# ---------------------------------------------------------------------------

class TestContextBundleTokenBudget:
    def test_budget_none_returns_full_source(self, tmp_path):
        """Without a budget, source is always included."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(repo, symbol_id="engine.py::Engine#class", storage_path=storage)
        assert "error" not in result
        assert result.get("source", "") != ""

    def test_compact_strategy_strips_source(self, tmp_path):
        """budget_strategy='compact' strips source bodies from all entries."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_ids=["engine.py::Engine#class", "utils.py::format_date#function"],
            token_budget=10000,
            budget_strategy="compact",
            storage_path=storage,
        )
        assert "error" not in result
        for sym in result["symbols"]:
            assert sym["source"] == "", f"Expected empty source, got: {sym['source'][:50]}"

    def test_compact_strategy_retains_signature(self, tmp_path):
        """compact mode keeps signature even when source is stripped."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_id="engine.py::Engine#class",
            token_budget=10000,
            budget_strategy="compact",
            storage_path=storage,
        )
        assert "error" not in result
        assert result.get("signature"), "Signature should be non-empty in compact mode"

    def test_tiny_budget_excludes_symbols(self, tmp_path):
        """A very small token_budget trims symbols that don't fit."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        # Engine class has substantial source; budget of 1 token should exclude everything
        result = get_context_bundle(
            repo,
            symbol_ids=["engine.py::Engine#class", "utils.py::format_date#function"],
            token_budget=1,
            budget_strategy="most_relevant",
            include_budget_report=True,
            storage_path=storage,
        )
        # Should not crash; budget_report should reflect exclusions
        assert "error" not in result or "budget_report" in result

    def test_budget_report_omitted_by_default(self, tmp_path):
        """budget_report is absent when include_budget_report=False (default)."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_id="engine.py::Engine#class",
            token_budget=10000,
            storage_path=storage,
        )
        assert "budget_report" not in result

    def test_budget_report_present_when_requested(self, tmp_path):
        """budget_report is present when include_budget_report=True."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_ids=["engine.py::Engine#class", "utils.py::format_date#function"],
            token_budget=10000,
            include_budget_report=True,
            storage_path=storage,
        )
        assert "error" not in result
        assert "budget_report" in result
        br = result["budget_report"]
        assert "budget_tokens" in br
        assert "used_tokens" in br
        assert "included_symbols" in br
        assert "excluded_symbols" in br
        assert "strategy" in br

    def test_budget_report_used_tokens_le_budget(self, tmp_path):
        """budget_report.used_tokens must not exceed token_budget."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_ids=["engine.py::Engine#class", "utils.py::format_date#function"],
            token_budget=200,
            include_budget_report=True,
            storage_path=storage,
        )
        assert "error" not in result
        br = result["budget_report"]
        assert br["used_tokens"] <= br["budget_tokens"]

    def test_invalid_budget_strategy_returns_error(self, tmp_path):
        """Unknown budget_strategy returns a structured error."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(
            repo,
            symbol_id="engine.py::Engine#class",
            token_budget=1000,
            budget_strategy="magic",
            storage_path=storage,
        )
        assert "error" in result
        assert "budget_strategy" in result["error"]

    def test_no_budget_backward_compat(self, tmp_path):
        """Without token_budget, response shape is unchanged (backward compat)."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_context_bundle(repo, symbol_id="utils.py::format_date#function", storage_path=storage)
        assert "error" not in result
        assert "symbol_id" in result
        assert "source" in result
        assert "budget_report" not in result


# ---------------------------------------------------------------------------
# get_ranked_context
# ---------------------------------------------------------------------------

class TestGetRankedContext:
    def test_returns_context_items(self, tmp_path):
        """Basic call returns context_items list."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="engine run", storage_path=storage)
        assert "error" not in result
        assert "context_items" in result
        assert isinstance(result["context_items"], list)

    def test_total_tokens_le_budget(self, tmp_path):
        """total_tokens must not exceed token_budget."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        budget = 100
        result = get_ranked_context(repo, query="engine", token_budget=budget, storage_path=storage)
        assert "error" not in result
        assert result["total_tokens"] <= budget

    def test_items_include_source(self, tmp_path):
        """Each context item includes a non-empty source field."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="Engine", token_budget=4000, storage_path=storage)
        assert "error" not in result
        for item in result["context_items"]:
            assert "source" in item

    def test_items_have_score_fields(self, tmp_path):
        """Each context item has relevance_score, centrality_score, combined_score."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="Engine", token_budget=4000, storage_path=storage)
        assert "error" not in result
        for item in result["context_items"]:
            assert "relevance_score" in item
            assert "centrality_score" in item
            assert "combined_score" in item
            assert "tokens" in item

    def test_bm25_strategy(self, tmp_path):
        """strategy='bm25' returns results without error."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="format date", strategy="bm25", storage_path=storage)
        assert "error" not in result
        assert "context_items" in result

    def test_budget_zero_returns_error(self, tmp_path):
        """token_budget=0 returns a structured error."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="engine", token_budget=0, storage_path=storage)
        assert "error" in result

    def test_include_kinds_filter(self, tmp_path):
        """include_kinds restricts results to specified kinds."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(
            repo, query="engine", token_budget=4000,
            include_kinds=["class"],
            storage_path=storage,
        )
        assert "error" not in result
        for item in result["context_items"]:
            # symbol_id encodes kind; verify via cross-check below
            # Just check we got results without crashing
            assert "symbol_id" in item

    def test_unknown_repo_returns_error(self, tmp_path):
        """Non-existent repo returns a structured error."""
        storage = str(tmp_path / ".index")
        result = get_ranked_context("no_such_repo", query="engine", storage_path=storage)
        assert "error" in result

    def test_query_too_long_returns_error(self, tmp_path):
        """Query exceeding 500 chars returns a structured error."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="x" * 501, storage_path=storage)
        assert "error" in result

    def test_meta_fields_present(self, tmp_path):
        """Response includes _meta with timing and savings fields."""
        repo, storage = _make_repo(tmp_path, _SMALL_REPO)
        result = get_ranked_context(repo, query="Engine", storage_path=storage)
        assert "error" not in result
        assert "_meta" in result
        meta = result["_meta"]
        assert "timing_ms" in meta
        assert "tokens_saved" in meta



"""Tests for Feature 8 — Optional Semantic / Embedding Search.

All tests run without any real embedding provider by mocking embed_texts.
The full no-provider error path and zero-perf-impact defaults are verified
without any optional dependencies installed.
"""

import pytest
from unittest.mock import patch

from jcodemunch_mcp.parser.symbols import Symbol
from jcodemunch_mcp.storage import IndexStore
from jcodemunch_mcp.tools.search_symbols import search_symbols, _cosine_similarity
from jcodemunch_mcp.tools.embed_repo import embed_repo, _detect_provider, _sym_text


# ── Helpers ──────────────────────────────────────────────────────────────────

def _make_symbol(id_, name, signature="", summary="", kind="function", file="src/a.py"):
    return Symbol(
        id=id_,
        file=file,
        name=name,
        qualified_name=name,
        kind=kind,
        language="python",
        signature=signature or f"def {name}():",
        byte_offset=0,
        byte_length=50,
        summary=summary,
    )


def _seed(tmp_path, symbols, raw_files=None):
    """Index a small synthetic repo and return (store, repo_id)."""
    store = IndexStore(base_path=str(tmp_path))
    if raw_files is None:
        raw_files = {s.file: f"def {s.name}(): pass\n" for s in symbols}
    file_languages = {s.file: "python" for s in symbols}
    source_files = list({s.file for s in symbols})
    store.save_index(
        owner="test",
        name="semantic",
        source_files=source_files,
        symbols=symbols,
        raw_files=raw_files,
        languages={"python": len(source_files)},
        file_languages=file_languages,
    )
    return store, "test/semantic"


# ── Unit: _cosine_similarity ─────────────────────────────────────────────────

def test_cosine_similarity_identical_vectors():
    v = [1.0, 0.0, 0.0]
    assert _cosine_similarity(v, v) == pytest.approx(1.0)


def test_cosine_similarity_orthogonal_vectors():
    assert _cosine_similarity([1.0, 0.0], [0.0, 1.0]) == pytest.approx(0.0)


def test_cosine_similarity_zero_vector_returns_zero():
    assert _cosine_similarity([0.0, 0.0], [1.0, 2.0]) == 0.0


def test_cosine_similarity_mismatched_lengths_returns_zero():
    assert _cosine_similarity([1.0, 2.0], [1.0]) == 0.0


def test_cosine_similarity_opposite_vectors():
    assert _cosine_similarity([1.0, 0.0], [-1.0, 0.0]) == pytest.approx(-1.0)


# ── Unit: _detect_provider ────────────────────────────────────────────────────

def test_detect_provider_none_when_nothing_set(monkeypatch):
    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    monkeypatch.delenv("GOOGLE_EMBED_MODEL", raising=False)
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    monkeypatch.delenv("OPENAI_EMBED_MODEL", raising=False)
    assert _detect_provider() is None


def test_detect_provider_sentence_transformers(monkeypatch):
    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    provider, model = _detect_provider()
    assert provider == "sentence_transformers"
    assert model == "all-MiniLM-L6-v2"


def test_detect_provider_gemini(monkeypatch):
    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/embedding-001")
    monkeypatch.delenv("OPENAI_EMBED_MODEL", raising=False)
    provider, model = _detect_provider()
    assert provider == "gemini"
    assert model == "models/embedding-001"


def test_detect_provider_openai(monkeypatch):
    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    monkeypatch.setenv("OPENAI_API_KEY", "sk-fake")
    monkeypatch.setenv("OPENAI_EMBED_MODEL", "text-embedding-3-small")
    provider, model = _detect_provider()
    assert provider == "openai"
    assert model == "text-embedding-3-small"


def test_detect_provider_sentence_transformers_wins_over_others(monkeypatch):
    """sentence-transformers takes priority when multiple providers are set."""
    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/embedding-001")
    provider, _ = _detect_provider()
    assert provider == "sentence_transformers"


def test_detect_provider_openai_requires_embed_model(monkeypatch):
    """OPENAI_API_KEY alone (used for local LLM) must NOT activate OpenAI embeddings."""
    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    monkeypatch.setenv("OPENAI_API_KEY", "local-llm")
    monkeypatch.delenv("OPENAI_EMBED_MODEL", raising=False)
    assert _detect_provider() is None


# ── Integration: semantic=False has zero impact ───────────────────────────────

def test_semantic_false_default_is_pure_bm25(tmp_path):
    """semantic=False (default) must not touch the embedding code at all."""
    symbols = [
        _make_symbol("s1", "connection_pool", summary="manages db connections"),
        _make_symbol("s2", "render_html", summary="renders html templates"),
    ]
    _seed(tmp_path, symbols)

    # No monkeypatching of providers — if any embedding code ran it would fail
    result = search_symbols("test/semantic", "connection_pool", storage_path=str(tmp_path))
    assert result["result_count"] >= 1
    assert result["results"][0]["name"] == "connection_pool"
    assert "_meta" in result
    assert "search_mode" not in result["_meta"]  # not set by BM25 path


# ── Integration: no provider → structured error ───────────────────────────────

def test_semantic_true_no_provider_returns_error(tmp_path, monkeypatch):
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    monkeypatch.delenv("GOOGLE_EMBED_MODEL", raising=False)
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    monkeypatch.delenv("OPENAI_EMBED_MODEL", raising=False)

    result = search_symbols("test/semantic", "foo", semantic=True, storage_path=str(tmp_path))
    assert result.get("error") == "no_embedding_provider"
    assert "message" in result


def test_embed_repo_no_provider_returns_error(tmp_path, monkeypatch):
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.delenv("GOOGLE_API_KEY", raising=False)
    monkeypatch.delenv("GOOGLE_EMBED_MODEL", raising=False)
    monkeypatch.delenv("OPENAI_API_KEY", raising=False)
    monkeypatch.delenv("OPENAI_EMBED_MODEL", raising=False)

    result = embed_repo("test/semantic", storage_path=str(tmp_path))
    assert result.get("error") == "no_embedding_provider"
    assert "message" in result


# ── Integration: mock provider — hybrid search ────────────────────────────────

def _make_vec(seed: float, dim: int = 8) -> list[float]:
    """Deterministic unit vector for testing."""
    import math
    v = [math.cos(seed + i * 0.5) for i in range(dim)]
    norm = math.sqrt(sum(x * x for x in v))
    return [x / norm for x in v]


def test_semantic_hybrid_surfaces_missed_bm25_symbol(tmp_path, monkeypatch):
    """Hybrid search returns connection_pool for 'database pool' even when BM25 misses."""
    symbols = [
        _make_symbol("s1", "connection_pool", summary="manages db connections"),
        _make_symbol("s2", "render_html", summary="renders html templates"),
        _make_symbol("s3", "parse_config", summary="parses config file"),
    ]
    store, repo = _seed(tmp_path, symbols)

    # Assign embeddings: connection_pool very similar to query, others distant
    query_vec = _make_vec(0.0)
    sym_vecs = {
        "s1": _make_vec(0.05),   # very close to query
        "s2": _make_vec(3.14),   # orthogonal-ish
        "s3": _make_vec(2.0),    # distant
    }

    def _mock_embed(texts, provider, model, task_type=None):
        if len(texts) == 1 and texts[0].startswith("database"):
            return [query_vec]
        # Called for symbol texts — map by index into sym order
        return [sym_vecs.get(f"s{i+1}", _make_vec(i)) for i in range(len(texts))]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    # Patch embed_texts at its definition site — both embed_repo and
    # _search_symbols_semantic import it from there at call time.
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        embed_repo(repo, storage_path=str(tmp_path))
        result = search_symbols(
            repo, "database pool", semantic=True, storage_path=str(tmp_path)
        )

    assert result.get("error") is None, result
    assert result["result_count"] >= 1
    names = [r["name"] for r in result["results"]]
    assert "connection_pool" in names, f"Expected connection_pool in {names}"
    assert result["_meta"]["search_mode"] == "hybrid"


def test_semantic_weight_zero_equals_pure_bm25(tmp_path, monkeypatch):
    """semantic_weight=0.0 must return identical ranking to pure BM25."""
    symbols = [
        _make_symbol("s1", "connection_pool", summary="database pool manager"),
        _make_symbol("s2", "render_html", summary="html renderer"),
    ]
    _seed(tmp_path, symbols)

    bm25_result = search_symbols(
        "test/semantic", "connection_pool", storage_path=str(tmp_path)
    )

    dummy_vec = _make_vec(1.0)

    def _mock_embed(texts, provider, model, task_type=None):
        return [dummy_vec for _ in texts]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        embed_repo("test/semantic", storage_path=str(tmp_path))
        sem_result = search_symbols(
            "test/semantic", "connection_pool",
            semantic=True, semantic_weight=0.0,
            storage_path=str(tmp_path),
        )

    # Both must rank connection_pool first
    assert bm25_result["results"][0]["name"] == "connection_pool"
    assert sem_result["results"][0]["name"] == "connection_pool"


def test_semantic_only_skips_bm25(tmp_path, monkeypatch):
    """semantic_only=True must surface symbols with no BM25 token overlap."""
    symbols = [
        _make_symbol("s1", "connection_pool", summary="manages db connections"),
        _make_symbol("s2", "render_html", summary="renders html"),
    ]
    _seed(tmp_path, symbols)

    # connection_pool gets a near-perfect cosine match; render_html is distant
    query_vec = _make_vec(0.0)
    sym_vecs = {"s1": _make_vec(0.01), "s2": _make_vec(3.0)}

    def _mock_embed(texts, provider, model, task_type=None):
        if len(texts) == 1:
            return [query_vec]
        return [sym_vecs.get(f"s{i+1}", _make_vec(i)) for i in range(len(texts))]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        embed_repo("test/semantic", storage_path=str(tmp_path))
        result = search_symbols(
            "test/semantic", "xyzzy_no_match",
            semantic=True, semantic_only=True,
            storage_path=str(tmp_path),
        )

    assert result.get("error") is None, result
    assert result["_meta"]["search_mode"] == "semantic_only"
    assert result["result_count"] >= 1
    assert result["results"][0]["name"] == "connection_pool"


# ── Integration: embed_repo tool ──────────────────────────────────────────────

def test_embed_repo_caches_and_skips_on_second_call(tmp_path, monkeypatch):
    """embed_repo must not recompute embeddings that are already stored."""
    symbols = [_make_symbol("s1", "foo"), _make_symbol("s2", "bar")]
    _seed(tmp_path, symbols)

    call_count = {"n": 0}

    def _mock_embed(texts, provider, model, task_type=None):
        call_count["n"] += len(texts)
        return [_make_vec(i) for i in range(len(texts))]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        r1 = embed_repo("test/semantic", storage_path=str(tmp_path))
        r2 = embed_repo("test/semantic", storage_path=str(tmp_path))

    assert r1["symbols_embedded"] == 2
    assert r2["symbols_embedded"] == 0
    assert r2.get("cached") is True
    # embed_texts was only called for the first invocation
    assert call_count["n"] == 2


def test_embed_repo_force_recomputes(tmp_path, monkeypatch):
    """force=True must recompute even when embeddings already exist."""
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    call_count = {"n": 0}

    def _mock_embed(texts, provider, model, task_type=None):
        call_count["n"] += len(texts)
        return [_make_vec(i) for i in range(len(texts))]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        embed_repo("test/semantic", storage_path=str(tmp_path))
        r2 = embed_repo("test/semantic", force=True, storage_path=str(tmp_path))

    assert r2["symbols_embedded"] == 1
    assert call_count["n"] == 2  # called twice total


def test_embed_repo_returns_stats(tmp_path, monkeypatch):
    """embed_repo result must include provider, model, dimension, and timing."""
    symbols = [_make_symbol("s1", "foo"), _make_symbol("s2", "bar")]
    _seed(tmp_path, symbols)

    def _mock_embed(texts, provider, model, task_type=None):
        return [_make_vec(i, dim=4) for i in range(len(texts))]

    monkeypatch.setenv("JCODEMUNCH_EMBED_MODEL", "all-MiniLM-L6-v2")
    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        result = embed_repo("test/semantic", storage_path=str(tmp_path))

    assert result["provider"] == "sentence_transformers"
    assert result["model"] == "all-MiniLM-L6-v2"
    assert result["symbols_total"] == 2
    assert result["symbols_embedded"] == 2
    assert result["embedding_dimension"] == 4
    assert "_meta" in result
    assert result["_meta"]["timing_ms"] >= 0


# ── Integration: EmbeddingStore ───────────────────────────────────────────────

def test_embedding_store_persist_and_retrieve(tmp_path):
    """Embeddings must survive a process restart (new EmbeddingStore instance)."""
    from jcodemunch_mcp.storage.embedding_store import EmbeddingStore

    symbols = [_make_symbol("s1", "foo")]
    store, _ = _seed(tmp_path, symbols)
    db_path = store._sqlite._db_path("test", "semantic")

    vec = [0.1, 0.2, 0.3, 0.4]
    es1 = EmbeddingStore(db_path)
    es1.set_many({"s1": vec})
    es1.set_dimension(4, "test-model")

    # New instance simulates a new process
    es2 = EmbeddingStore(db_path)
    retrieved = es2.get("s1")
    assert retrieved is not None
    assert len(retrieved) == 4
    for a, b in zip(retrieved, vec):
        assert abs(a - b) < 1e-5
    assert es2.get_dimension() == 4


def test_embedding_store_delete_many(tmp_path):
    from jcodemunch_mcp.storage.embedding_store import EmbeddingStore

    symbols = [_make_symbol("s1", "foo"), _make_symbol("s2", "bar")]
    store, _ = _seed(tmp_path, symbols)
    db_path = store._sqlite._db_path("test", "semantic")

    es = EmbeddingStore(db_path)
    es.set_many({"s1": [1.0, 0.0], "s2": [0.0, 1.0]})
    assert es.count() == 2

    es.delete_many(["s1"])
    assert es.count() == 1
    assert es.get("s1") is None
    assert es.get("s2") is not None


# ── Unit: task-type helpers ───────────────────────────────────────────────────

def test_gemini_task_aware_default_is_on(monkeypatch):
    """Task-type support is enabled by default (no env var set)."""
    from jcodemunch_mcp.tools.embed_repo import _gemini_task_aware
    monkeypatch.delenv("GEMINI_EMBED_TASK_AWARE", raising=False)
    assert _gemini_task_aware() is True


def test_gemini_task_aware_opt_out(monkeypatch):
    """GEMINI_EMBED_TASK_AWARE=0 disables task types."""
    from jcodemunch_mcp.tools.embed_repo import _gemini_task_aware
    for val in ("0", "false", "no", "off"):
        monkeypatch.setenv("GEMINI_EMBED_TASK_AWARE", val)
        assert _gemini_task_aware() is False, f"should be False for {val!r}"


# ── Integration: Gemini task-type routing ─────────────────────────────────────

def test_embed_repo_passes_retrieval_document_for_gemini(tmp_path, monkeypatch):
    """embed_repo must pass task_type='RETRIEVAL_DOCUMENT' when using Gemini."""
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    received_task_types: list = []

    def _mock_embed(texts, provider, model, task_type=None):
        received_task_types.append(task_type)
        return [_make_vec(0)]

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/text-embedding-004")
    monkeypatch.delenv("GEMINI_EMBED_TASK_AWARE", raising=False)

    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        result = embed_repo("test/semantic", storage_path=str(tmp_path))

    assert result.get("error") is None, result
    assert all(tt == "RETRIEVAL_DOCUMENT" for tt in received_task_types), received_task_types
    assert result.get("task_type") == "RETRIEVAL_DOCUMENT"


def test_search_symbols_passes_code_retrieval_query_for_gemini(tmp_path, monkeypatch):
    """semantic search must use CODE_RETRIEVAL_QUERY for the query and
    RETRIEVAL_DOCUMENT for lazy symbol embedding when using Gemini."""
    symbols = [_make_symbol("s1", "authenticate_user", summary="handles auth")]
    _seed(tmp_path, symbols)

    calls: list[dict] = []
    query_vec = _make_vec(0.0)
    sym_vec = _make_vec(0.01)

    def _mock_embed(texts, provider, model, task_type=None):
        calls.append({"texts": texts, "task_type": task_type})
        # First call is the query; subsequent calls are symbol batches
        if len(texts) == 1 and texts[0] == "authenticate user":
            return [query_vec]
        return [sym_vec for _ in texts]

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/text-embedding-004")
    monkeypatch.delenv("GEMINI_EMBED_TASK_AWARE", raising=False)

    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        result = search_symbols(
            "test/semantic", "authenticate user",
            semantic=True, storage_path=str(tmp_path),
        )

    assert result.get("error") is None, result
    query_calls = [c for c in calls if c["texts"] == ["authenticate user"]]
    symbol_calls = [c for c in calls if c["texts"] != ["authenticate user"]]
    assert query_calls, "expected at least one query embedding call"
    assert all(c["task_type"] == "CODE_RETRIEVAL_QUERY" for c in query_calls)
    assert all(c["task_type"] == "RETRIEVAL_DOCUMENT" for c in symbol_calls)


def test_task_type_opt_out_disables_task_types(tmp_path, monkeypatch):
    """GEMINI_EMBED_TASK_AWARE=0 must result in no task_type being passed."""
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    received: list = []

    def _mock_embed(texts, provider, model, task_type=None):
        received.append(task_type)
        return [_make_vec(0)]

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/text-embedding-004")
    monkeypatch.setenv("GEMINI_EMBED_TASK_AWARE", "0")

    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        result = embed_repo("test/semantic", storage_path=str(tmp_path))

    assert result.get("error") is None, result
    assert all(tt is None for tt in received), received
    assert "task_type" not in result


def test_task_type_change_triggers_reembed(tmp_path, monkeypatch):
    """Toggling GEMINI_EMBED_TASK_AWARE must invalidate existing embeddings."""
    symbols = [_make_symbol("s1", "foo")]
    _seed(tmp_path, symbols)

    call_count = {"n": 0}

    def _mock_embed(texts, provider, model, task_type=None):
        call_count["n"] += len(texts)
        return [_make_vec(0)]

    monkeypatch.delenv("JCODEMUNCH_EMBED_MODEL", raising=False)
    monkeypatch.setenv("GOOGLE_API_KEY", "fake-key")
    monkeypatch.setenv("GOOGLE_EMBED_MODEL", "models/text-embedding-004")

    with patch("jcodemunch_mcp.tools.embed_repo.embed_texts", side_effect=_mock_embed):
        # First run: task-type aware (RETRIEVAL_DOCUMENT stored)
        monkeypatch.delenv("GEMINI_EMBED_TASK_AWARE", raising=False)
        embed_repo("test/semantic", storage_path=str(tmp_path))
        assert call_count["n"] == 1

        # Second run: same settings — should skip (all embedded)
        embed_repo("test/semantic", storage_path=str(tmp_path))
        assert call_count["n"] == 1  # no new calls

        # Third run: opt-out — task_type changed → must re-embed
        monkeypatch.setenv("GEMINI_EMBED_TASK_AWARE", "0")
        r = embed_repo("test/semantic", storage_path=str(tmp_path))
        assert r["symbols_embedded"] == 1
        assert call_count["n"] == 2


# ── Integration: EmbeddingStore task type persistence ─────────────────────────

def test_embedding_store_task_type_persist_and_retrieve(tmp_path):
    """set_task_type / get_task_type must round-trip through SQLite."""
    from jcodemunch_mcp.storage.embedding_store import EmbeddingStore

    symbols = [_make_symbol("s1", "foo")]
    store, _ = _seed(tmp_path, symbols)
    db_path = store._sqlite._db_path("test", "semantic")

    es1 = EmbeddingStore(db_path)
    assert es1.get_task_type() is None  # not set yet

    es1.set_task_type("RETRIEVAL_DOCUMENT")

    es2 = EmbeddingStore(db_path)  # new instance simulates new process
    assert es2.get_task_type() == "RETRIEVAL_DOCUMENT"

    es2.set_task_type("")  # clearing
    es3 = EmbeddingStore(db_path)
    assert es3.get_task_type() == ""


# ── Unit: _normalise_gemini_task_type ─────────────────────────────────────────

class _FakeTaskType:
    """Minimal stand-in for genai.protos.TaskType enum entries."""
    def __init__(self, name):
        self.name = name


class _FakeGenai:
    """Stand-in for google.generativeai module with a limited TaskType enum."""
    class protos:
        TaskType = [
            _FakeTaskType("RETRIEVAL_QUERY"),
            _FakeTaskType("RETRIEVAL_DOCUMENT"),
        ]


class _NewGenai:
    """Stand-in for a newer SDK that also has CODE_RETRIEVAL_QUERY."""
    class protos:
        TaskType = [
            _FakeTaskType("RETRIEVAL_QUERY"),
            _FakeTaskType("RETRIEVAL_DOCUMENT"),
            _FakeTaskType("CODE_RETRIEVAL_QUERY"),
        ]


def test_normalise_gemini_task_type_none_passthrough():
    from jcodemunch_mcp.tools.embed_repo import _normalise_gemini_task_type
    assert _normalise_gemini_task_type(_FakeGenai, None) is None


def test_normalise_gemini_task_type_supported_passes_through():
    """RETRIEVAL_DOCUMENT is in both old and new SDK — should return as-is."""
    from jcodemunch_mcp.tools.embed_repo import _normalise_gemini_task_type
    assert _normalise_gemini_task_type(_FakeGenai, "RETRIEVAL_DOCUMENT") == "RETRIEVAL_DOCUMENT"


def test_normalise_gemini_task_type_code_retrieval_falls_back_on_old_sdk():
    """CODE_RETRIEVAL_QUERY absent from old SDK → falls back to RETRIEVAL_QUERY."""
    from jcodemunch_mcp.tools.embed_repo import _normalise_gemini_task_type
    result = _normalise_gemini_task_type(_FakeGenai, "CODE_RETRIEVAL_QUERY")
    assert result == "RETRIEVAL_QUERY"


def test_normalise_gemini_task_type_code_retrieval_passes_on_new_sdk():
    """CODE_RETRIEVAL_QUERY present in new SDK → returned unchanged."""
    from jcodemunch_mcp.tools.embed_repo import _normalise_gemini_task_type
    result = _normalise_gemini_task_type(_NewGenai, "CODE_RETRIEVAL_QUERY")
    assert result == "CODE_RETRIEVAL_QUERY"


def test_normalise_gemini_task_type_unsupported_no_fallback_returns_none():
    """Unknown task type with no fallback entry → None (omit from API call)."""
    from jcodemunch_mcp.tools.embed_repo import _normalise_gemini_task_type
    result = _normalise_gemini_task_type(_FakeGenai, "CLUSTERING")
    assert result is None
